# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.spdx3.model import Hash, HashAlgorithm
from spdx_tools.spdx.model.checksum import Checksum as Spdx2_Checksum
from spdx_tools.spdx.model.checksum import ChecksumAlgorithm


def bump_checksum(spdx2_checksum: Spdx2_Checksum) -> Hash:
    algorithm: HashAlgorithm = convert_checksum_algorithm_to_hash_algorithm(spdx2_checksum.algorithm)
    value: str = spdx2_checksum.value

    return Hash(algorithm, value)


def convert_checksum_algorithm_to_hash_algorithm(checksum_algorithm: ChecksumAlgorithm) -> HashAlgorithm:
    if checksum_algorithm.name.startswith("BLAKE"):
        return HashAlgorithm[checksum_algorithm.name.replace("_", "")]
    if checksum_algorithm == ChecksumAlgorithm.ADLER32:
        return HashAlgorithm.OTHER
    return HashAlgorithm[checksum_algorithm.name]



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) spdx contributors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from spdx_tools.spdx.formats import FileFormat, file_name_to_format
from spdx_tools.spdx.parser.json import json_parser
from spdx_tools.spdx.parser.rdf import rdf_parser
from spdx_tools.spdx.parser.tagvalue import tagvalue_parser
from spdx_tools.spdx.parser.xml import xml_parser
from spdx_tools.spdx.parser.yaml import yaml_parser


def parse_file(file_name: str, encoding: str = "utf-8"):
    if encoding != "utf-8":
        logging.warning(
            "It's recommended to use the UTF-8 encoding for any SPDX file. Consider changing the encoding of the file."
        )

    input_format = file_name_to_format(file_name)
    if input_format == FileFormat.RDF_XML:
        return rdf_parser.parse_from_file(file_name, encoding)
    elif input_format == FileFormat.TAG_VALUE:
        return tagvalue_parser.parse_from_file(file_name, encoding)
    elif input_format == FileFormat.JSON:
        return json_parser.parse_from_file(file_name, encoding)
    elif input_format == FileFormat.XML:
        return xml_parser.parse_from_file(file_name, encoding)
    elif input_format == FileFormat.YAML:
        return yaml_parser.parse_from_file(file_name, encoding)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List

from spdx_tools.spdx.model import Document, Package, Relationship, RelationshipType


def find_package_contains_file_relationships(document: Document, package: Package) -> List[Relationship]:
    file_ids_in_document = [file.spdx_id for file in document.files]
    package_contains_relationships = filter_by_type_and_origin(
        document.relationships, RelationshipType.CONTAINS, package.spdx_id
    )
    return [
        relationship
        for relationship in package_contains_relationships
        if relationship.related_spdx_element_id in file_ids_in_document
    ]


def find_file_contained_by_package_relationships(document: Document, package: Package) -> List[Relationship]:
    file_ids_in_document = [file.spdx_id for file in document.files]
    contained_by_package_relationships = filter_by_type_and_target(
        document.relationships, RelationshipType.CONTAINED_BY, package.spdx_id
    )
    return [
        relationship
        for relationship in contained_by_package_relationships
        if relationship.spdx_element_id in file_ids_in_document
    ]


def filter_by_type_and_target(
    relationships: List[Relationship], relationship_type: RelationshipType, target_id: str
) -> List[Relationship]:
    return [
        relationship
        for relationship in relationships
        if relationship.relationship_type == relationship_type and relationship.related_spdx_element_id == target_id
    ]


def filter_by_type_and_origin(
    relationships: List[Relationship], relationship_type: RelationshipType, origin_id: str
) -> List[Relationship]:
    return [
        relationship
        for relationship in relationships
        if relationship.relationship_type == relationship_type and relationship.spdx_element_id == origin_id
    ]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model.licensing.any_license_info import AnyLicenseInfo
from spdx_tools.spdx3.model.licensing.license import License
from spdx_tools.spdx3.model.licensing.license_addition import LicenseAddition


@dataclass_with_properties
class WithAdditionOperator(AnyLicenseInfo):
    subject_license: License
    subject_addition: LicenseAddition

    def __init__(self, subject_license: License, subject_addition: LicenseAddition):
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest
from license_expression import LicenseExpression

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx3.bump_from_spdx2.license_expression import (
    bump_license_expression,
    bump_license_expression_or_none_or_no_assertion,
)
from spdx_tools.spdx3.model.licensing import (
    ConjunctiveLicenseSet,
    CustomLicense,
    CustomLicenseAddition,
    DisjunctiveLicenseSet,
    ListedLicense,
    ListedLicenseException,
    NoAssertionLicense,
    NoneLicense,
    WithAdditionOperator,
)
from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from tests.spdx.fixtures import extracted_licensing_info_fixture


@pytest.mark.parametrize(
    "element, expected_class",
    [
        (SpdxNoAssertion(), NoAssertionLicense),
        (SpdxNone(), NoneLicense),
        (spdx_licensing.parse("MIT"), ListedLicense),
    ],
)
def test_license_expression_or_none_or_no_assertion(element, expected_class):
    license_info = bump_license_expression_or_none_or_no_assertion(element, [])

    assert isinstance(license_info, expected_class)


@pytest.mark.parametrize(
    "license_expression, extracted_licensing_info, expected_element",
    [
        (spdx_licensing.parse("MIT"), [], ListedLicense("MIT", "MIT", "blank")),
        (spdx_licensing.parse("LGPL-2.0"), [], ListedLicense("LGPL-2.0-only", "LGPL-2.0-only", "blank")),
        (
            spdx_licensing.parse("LicenseRef-1"),
            [extracted_licensing_info_fixture()],
            CustomLicense("LicenseRef-1", "licenseName", "extractedText"),
        ),
        (
            spdx_licensing.parse("MIT AND LGPL-2.0"),
            [],
            ConjunctiveLicenseSet(
                [ListedLicense("MIT", "MIT", "blank"), ListedLicense("LGPL-2.0-only", "LGPL-2.0-only", "blank")]
            ),
        ),
        (
            spdx_licensing.parse("LicenseRef-1 OR LGPL-2.0"),
            [extracted_licensing_info_fixture()],
            DisjunctiveLicenseSet(
                [
                    CustomLicense("LicenseRef-1", "licenseName", "extractedText"),
                    ListedLicense("LGPL-2.0-only", "LGPL-2.0-only", "blank"),
                ]
            ),
        ),
        (
            spdx_licensing.parse("LGPL-2.0 WITH 389-exception"),
            [],
            WithAdditionOperator(
                ListedLicense("LGPL-2.0-only", "LGPL-2.0-only", "blank"),
                ListedLicenseException("389-exception", "", ""),
            ),
        ),
        (
            spdx_licensing.parse("LicenseRef-1 WITH custom-exception"),
            [
                extracted_licensing_info_fixture(),
                extracted_licensing_info_fixture("custom-exception", "This is a custom exception", "exceptionName"),
            ],
            WithAdditionOperator(
                CustomLicense("LicenseRef-1", "licenseName", "extractedText"),
                CustomLicenseAddition("custom-exception", "exceptionName", "This is a custom exception"),
            ),
        ),
        (
            spdx_licensing.parse("MIT AND LicenseRef-1 WITH custom-exception"),
            [
                extracted_licensing_info_fixture(),
                extracted_licensing_info_fixture("custom-exception", "This is a custom exception", "exceptionName"),
            ],
            ConjunctiveLicenseSet(
                [
                    ListedLicense("MIT", "MIT", "blank"),
                    WithAdditionOperator(
                        CustomLicense("LicenseRef-1", "licenseName", "extractedText"),
                        CustomLicenseAddition("custom-exception", "exceptionName", "This is a custom exception"),
                    ),
                ]
            ),
        ),
    ],
)
def test_license_expression_bump(license_expression: LicenseExpression, extracted_licensing_info, expected_element):
    license_info = bump_license_expression(license_expression, extracted_licensing_info)

    assert license_info == expected_element



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
from enum import Enum, auto

from beartype.typing import Any, Optional


class SpdxElementType(Enum):
    LICENSE_EXPRESSION = auto()
    PACKAGE_VERIFICATION_CODE = auto()
    EXTERNAL_DOCUMENT_REF = auto()
    CHECKSUM = auto()
    EXTERNAL_PACKAGE_REF = auto()
    ACTOR = auto()
    DOCUMENT = auto()
    CREATION_INFO = auto()
    PACKAGE = auto()
    FILE = auto()
    SNIPPET = auto()
    ANNOTATION = auto()
    RELATIONSHIP = auto()
    EXTRACTED_LICENSING_INFO = auto()


@dataclass(frozen=True)
class ValidationContext:
    spdx_id: Optional[str] = None  # not every type has an id, or it might be missing
    parent_id: Optional[str] = None  # if a parent is known and has a valid id
    element_type: Optional[SpdxElementType] = None
    full_element: Any = None  # can be any class of the data model


@dataclass(frozen=True)
class ValidationMessage:
    validation_message: str
    context: ValidationContext



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0


import re
from re import Pattern


class Version:
    VERSION_REGEX: Pattern = re.compile(r"^(\d+)\.(\d+)$")

    major: int
    minor: int

    @classmethod
    def is_valid_version_string(cls, value: str) -> bool:
        return cls.VERSION_REGEX.match(value) is not None

    # No type hint for Python reasons.
    # See https://stackoverflow.com/questions/33533148/how-do-i-type-hint-a-method-with-the-type-of-the-enclosing-class
    @classmethod
    def from_string(cls, value: str):
        if not Version.is_valid_version_string(value):
            raise ValueError(f"{value} is not a valid version string")

        match = cls.VERSION_REGEX.match(value)
        return cls(int(match.group(1)), int(match.group(2)))

    def __init__(self, major: int, minor: int):
        self.major = major
        self.minor = minor

    def __str__(self):
        return f"{self.major}.{self.minor}"

    def __eq__(self, other):
        if not isinstance(other, Version):
            return False
        return self.major == other.major and self.minor == other.minor



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict
from rdflib import RDF, RDFS, BNode, Graph, Literal, URIRef

from spdx_tools.spdx.casing_tools import snake_case_to_camel_case
from spdx_tools.spdx.model import Relationship, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.writer_utils import add_namespace_to_spdx_id


def add_relationship_to_graph(
    relationship: Relationship, graph: Graph, doc_namespace: str, external_doc_ref_to_namespace: Dict[str, str]
):
    relationship_node = BNode()
    graph.add((relationship_node, RDF.type, SPDX_NAMESPACE.Relationship))
    graph.add(
        (
            relationship_node,
            SPDX_NAMESPACE.relationshipType,
            SPDX_NAMESPACE[f"relationshipType_{snake_case_to_camel_case(relationship.relationship_type.name)}"],
        )
    )
    if isinstance(relationship.related_spdx_element_id, SpdxNone):
        graph.add((relationship_node, SPDX_NAMESPACE.relatedSpdxElement, SPDX_NAMESPACE.none))
    elif isinstance(relationship.related_spdx_element_id, SpdxNoAssertion):
        graph.add((relationship_node, SPDX_NAMESPACE.relatedSpdxElement, SPDX_NAMESPACE.noassertion))
    else:
        graph.add(
            (
                relationship_node,
                SPDX_NAMESPACE.relatedSpdxElement,
                URIRef(
                    add_namespace_to_spdx_id(
                        relationship.related_spdx_element_id, doc_namespace, external_doc_ref_to_namespace
                    )
                ),
            )
        )
    if relationship.comment:
        graph.add((relationship_node, RDFS.comment, Literal(relationship.comment)))
    relationship_resource = URIRef(
        add_namespace_to_spdx_id(relationship.spdx_element_id, doc_namespace, external_doc_ref_to_namespace)
    )
    graph.add((relationship_resource, SPDX_NAMESPACE.relationship, relationship_node))



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import Bundle
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.spdx_collection_writer import write_collection


def write_bundle(bundle: Bundle, text_output: TextIO, heading: bool = True):
    if heading:
        text_output.write("## Bundle\n")
    write_collection(bundle, text_output)
    write_value("context", bundle.context, text_output)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict, List, Optional, Union
from license_expression import LicenseExpression

from spdx_tools.spdx.model import Checksum, File, FileType, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.parser.jsonlikedict.checksum_parser import ChecksumParser
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import (
    parse_field_or_log_error,
    parse_field_or_no_assertion_or_none,
)
from spdx_tools.spdx.parser.jsonlikedict.license_expression_parser import LicenseExpressionParser
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)


class FileParser:
    logger: Logger
    checksum_parser: ChecksumParser
    license_expression_parser: LicenseExpressionParser

    def __init__(self):
        self.logger = Logger()
        self.checksum_parser = ChecksumParser()
        self.license_expression_parser = LicenseExpressionParser()

    def parse_file(self, file_dict: Dict) -> Optional[File]:
        logger = Logger()
        name: Optional[str] = file_dict.get("fileName")
        spdx_id: Optional[str] = file_dict.get("SPDXID")
        checksums_list: List[Dict] = file_dict.get("checksums")
        checksums: List[Checksum] = parse_field_or_log_error(
            logger, checksums_list, self.checksum_parser.parse_checksum, field_is_list=True
        )

        attribution_texts: List[str] = file_dict.get("attributionTexts", [])
        comment: Optional[str] = file_dict.get("comment")
        copyright_text: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = parse_field_or_no_assertion_or_none(
            file_dict.get("copyrightText")
        )
        file_contributors: List[str] = file_dict.get("fileContributors", [])
        file_types: List[FileType] = parse_field_or_log_error(
            logger, file_dict.get("fileTypes"), self.parse_file_types
        )

        license_comments: Optional[str] = file_dict.get("licenseComments")

        license_concluded: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = parse_field_or_log_error(
            logger, file_dict.get("licenseConcluded"), self.license_expression_parser.parse_license_expression
        )

        license_info_in_files: List[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = parse_field_or_log_error(
            logger,
            file_dict.get("licenseInfoInFiles"),
            self.license_expression_parser.parse_license_expression,
            field_is_list=True,
        )
        notice_text: Optional[str] = file_dict.get("noticeText")
        raise_parsing_error_if_logger_has_messages(logger, "File")

        file = construct_or_raise_parsing_error(
            File,
            dict(
                name=name,
                spdx_id=spdx_id,
                checksums=checksums,
                attribution_texts=attribution_texts,
                comment=comment,
                copyright_text=copyright_text,
                file_types=file_types,
                contributors=file_contributors,
                license_comment=license_comments,
                license_concluded=license_concluded,
                license_info_in_file=license_info_in_files,
                notice=notice_text,
            ),
        )
        return file

    @staticmethod
    def parse_file_types(file_types_list: List[str]) -> List[FileType]:
        logger = Logger()
        file_types = []
        for file_type in file_types_list:
            try:
                file_type = FileType[file_type]
            except KeyError:
                logger.append(f"Invalid FileType: {file_type}")
                continue
            file_types.append(file_type)
        raise_parsing_error_if_logger_has_messages(logger, "FileType")
        return file_types



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Union
from unittest import mock
from unittest.mock import MagicMock, NonCallableMagicMock

import pytest

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.jsonschema.annotation_converter import AnnotationConverter
from spdx_tools.spdx.jsonschema.package_converter import PackageConverter
from spdx_tools.spdx.jsonschema.package_properties import PackageProperty
from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    Annotation,
    AnnotationType,
    Checksum,
    ChecksumAlgorithm,
    Document,
    Package,
    PackagePurpose,
    PackageVerificationCode,
    SpdxNoAssertion,
    SpdxNone,
)
from spdx_tools.spdx.model.spdx_no_assertion import SPDX_NO_ASSERTION_STRING
from spdx_tools.spdx.model.spdx_none import SPDX_NONE_STRING
from tests.spdx.fixtures import (
    annotation_fixture,
    creation_info_fixture,
    document_fixture,
    external_package_ref_fixture,
    package_fixture,
)
from tests.spdx.mock_utils import assert_mock_method_called_with_arguments


@pytest.fixture
@mock.patch("spdx_tools.spdx.jsonschema.checksum_converter.ChecksumConverter", autospec=True)
@mock.patch("spdx_tools.spdx.jsonschema.annotation_converter.AnnotationConverter", autospec=True)
@mock.patch(
    "spdx_tools.spdx.jsonschema.package_verification_code_converter.PackageVerificationCodeConverter", autospec=True
)
@mock.patch("spdx_tools.spdx.jsonschema.external_package_ref_converter.ExternalPackageRefConverter", autospec=True)
def converter(
    package_ref_converter_mock: MagicMock,
    verification_code_converter_mock: MagicMock,
    annotation_converter_mock: MagicMock,
    checksum_converter_mock: MagicMock,
) -> PackageConverter:
    converter = PackageConverter()
    converter.checksum_converter = checksum_converter_mock()
    converter.annotation_converter = annotation_converter_mock()
    converter.package_verification_code_converter = verification_code_converter_mock()
    converter.external_package_ref_converter = package_ref_converter_mock()
    return converter


@pytest.mark.parametrize(
    "external_package_ref_property,expected",
    [
        (PackageProperty.SPDX_ID, "SPDXID"),
        (PackageProperty.ANNOTATIONS, "annotations"),
        (PackageProperty.ATTRIBUTION_TEXTS, "attributionTexts"),
        (PackageProperty.BUILT_DATE, "builtDate"),
        (PackageProperty.CHECKSUMS, "checksums"),
        (PackageProperty.COMMENT, "comment"),
        (PackageProperty.COPYRIGHT_TEXT, "copyrightText"),
        (PackageProperty.DESCRIPTION, "description"),
        (PackageProperty.DOWNLOAD_LOCATION, "downloadLocation"),
        (PackageProperty.EXTERNAL_REFS, "externalRefs"),
        (PackageProperty.FILES_ANALYZED, "filesAnalyzed"),
        (PackageProperty.HOMEPAGE, "homepage"),
        (PackageProperty.LICENSE_COMMENTS, "licenseComments"),
        (PackageProperty.LICENSE_CONCLUDED, "licenseConcluded"),
        (PackageProperty.LICENSE_DECLARED, "licenseDeclared"),
        (PackageProperty.LICENSE_INFO_FROM_FILES, "licenseInfoFromFiles"),
        (PackageProperty.NAME, "name"),
        (PackageProperty.ORIGINATOR, "originator"),
        (PackageProperty.PACKAGE_FILE_NAME, "packageFileName"),
        (PackageProperty.PACKAGE_VERIFICATION_CODE, "packageVerificationCode"),
        (PackageProperty.PRIMARY_PACKAGE_PURPOSE, "primaryPackagePurpose"),
        (PackageProperty.RELEASE_DATE, "releaseDate"),
        (PackageProperty.SOURCE_INFO, "sourceInfo"),
        (PackageProperty.SUMMARY, "summary"),
        (PackageProperty.SUPPLIER, "supplier"),
        (PackageProperty.VALID_UNTIL_DATE, "validUntilDate"),
        (PackageProperty.VERSION_INFO, "versionInfo"),
    ],
)
def test_json_property_names(
    converter: PackageConverter, external_package_ref_property: PackageProperty, expected: str
):
    assert converter.json_property_name(external_package_ref_property) == expected


def test_json_type(converter: PackageConverter):
    assert converter.get_json_type() == PackageProperty


def test_data_model_type(converter: PackageConverter):
    assert converter.get_data_model_type() == Package


def test_successful_conversion(converter: PackageConverter):
    converter.checksum_converter.convert.return_value = "mock_converted_checksum"
    converter.annotation_converter.convert.return_value = "mock_converted_annotation"
    converter.package_verification_code_converter.convert.return_value = "mock_converted_verification_code"
    converter.external_package_ref_converter.convert.return_value = "mock_package_ref"
    package = Package(
        spdx_id="packageId",
        name="name",
        download_location="downloadLocation",
        version="version",
        file_name="fileName",
        supplier=Actor(ActorType.PERSON, "supplierName"),
        originator=Actor(ActorType.PERSON, "originatorName"),
        files_analyzed=True,
        verification_code=PackageVerificationCode("value"),
        checksums=[Checksum(ChecksumAlgorithm.SHA1, "sha1"), Checksum(ChecksumAlgorithm.BLAKE2B_256, "blake")],
        homepage="homepage",
        source_info="sourceInfo",
        license_concluded=spdx_licensing.parse("MIT and GPL-2.0"),
        license_info_from_files=[spdx_licensing.parse("MIT"), spdx_licensing.parse("GPL-2.0")],
        license_declared=spdx_licensing.parse("MIT or GPL-2.0 "),
        license_comment="licenseComment",
        copyright_text="copyrightText",
        summary="summary",
        description="description",
        comment="comment",
        external_references=[external_package_ref_fixture()],
        attribution_texts=["attributionText1", "attributionText2"],
        primary_package_purpose=PackagePurpose.APPLICATION,
        release_date=datetime(2022, 12, 1),
        built_date=datetime(2022, 12, 2),
        valid_until_date=datetime(2022, 12, 3),
    )

    annotation = Annotation(
        package.spdx_id,
        AnnotationType.REVIEW,
        Actor(ActorType.TOOL, "toolName"),
        datetime(2022, 12, 5),
        "review comment",
    )
    document = Document(creation_info_fixture(), packages=[package], annotations=[annotation])

    converted_dict = converter.convert(package, document)

    assert converted_dict == {
        converter.json_property_name(PackageProperty.SPDX_ID): "packageId",
        converter.json_property_name(PackageProperty.ANNOTATIONS): ["mock_converted_annotation"],
        converter.json_property_name(PackageProperty.ATTRIBUTION_TEXTS): ["attributionText1", "attributionText2"],
        converter.json_property_name(PackageProperty.NAME): "name",
        converter.json_property_name(PackageProperty.DOWNLOAD_LOCATION): "downloadLocation",
        converter.json_property_name(PackageProperty.VERSION_INFO): "version",
        converter.json_property_name(PackageProperty.PACKAGE_FILE_NAME): "fileName",
        converter.json_property_name(PackageProperty.SUPPLIER): "Person: supplierName",
        converter.json_property_name(PackageProperty.ORIGINATOR): "Person: originatorName",
        converter.json_property_name(PackageProperty.FILES_ANALYZED): True,
        converter.json_property_name(PackageProperty.PACKAGE_VERIFICATION_CODE): "mock_converted_verification_code",
        converter.json_property_name(PackageProperty.CHECKSUMS): [
            "mock_converted_checksum",
            "mock_converted_checksum",
        ],
        converter.json_property_name(PackageProperty.HOMEPAGE): "homepage",
        converter.json_property_name(PackageProperty.SOURCE_INFO): "sourceInfo",
        converter.json_property_name(PackageProperty.LICENSE_CONCLUDED): "MIT AND GPL-2.0-only",
        converter.json_property_name(PackageProperty.LICENSE_INFO_FROM_FILES): ["MIT", "GPL-2.0-only"],
        converter.json_property_name(PackageProperty.LICENSE_DECLARED): "MIT OR GPL-2.0-only",
        converter.json_property_name(PackageProperty.LICENSE_COMMENTS): "licenseComment",
        converter.json_property_name(PackageProperty.COPYRIGHT_TEXT): "copyrightText",
        converter.json_property_name(PackageProperty.SUMMARY): "summary",
        converter.json_property_name(PackageProperty.DESCRIPTION): "description",
        converter.json_property_name(PackageProperty.COMMENT): "comment",
        converter.json_property_name(PackageProperty.EXTERNAL_REFS): ["mock_package_ref"],
        converter.json_property_name(PackageProperty.PRIMARY_PACKAGE_PURPOSE): "APPLICATION",
        converter.json_property_name(PackageProperty.RELEASE_DATE): "2022-12-01T00:00:00Z",
        converter.json_property_name(PackageProperty.BUILT_DATE): "2022-12-02T00:00:00Z",
        converter.json_property_name(PackageProperty.VALID_UNTIL_DATE): "2022-12-03T00:00:00Z",
    }


def test_null_values(converter: PackageConverter):
    package = package_fixture(
        built_date=None,
        release_date=None,
        valid_until_date=None,
        homepage=None,
        license_concluded=None,
        license_declared=None,
        originator=None,
        verification_code=None,
        primary_package_purpose=None,
        supplier=None,
        version=None,
        file_name=None,
        source_info=None,
        license_comment=None,
        copyright_text=None,
        summary=None,
        description=None,
        comment=None,
        attribution_texts=[],
        checksums=[],
        external_references=[],
        license_info_from_files=[],
    )

    document = Document(creation_info_fixture(), packages=[package])

    converted_dict = converter.convert(package, document)

    assert converter.json_property_name(PackageProperty.VERSION_INFO) not in converted_dict
    assert converter.json_property_name(PackageProperty.PACKAGE_FILE_NAME) not in converted_dict
    assert converter.json_property_name(PackageProperty.SUPPLIER) not in converted_dict
    assert converter.json_property_name(PackageProperty.ORIGINATOR) not in converted_dict
    assert converter.json_property_name(PackageProperty.PACKAGE_VERIFICATION_CODE) not in converted_dict
    assert converter.json_property_name(PackageProperty.HOMEPAGE) not in converted_dict
    assert converter.json_property_name(PackageProperty.SOURCE_INFO) not in converted_dict
    assert converter.json_property_name(PackageProperty.LICENSE_CONCLUDED) not in converted_dict
    assert converter.json_property_name(PackageProperty.LICENSE_DECLARED) not in converted_dict
    assert converter.json_property_name(PackageProperty.LICENSE_COMMENTS) not in converted_dict
    assert converter.json_property_name(PackageProperty.COPYRIGHT_TEXT) not in converted_dict
    assert converter.json_property_name(PackageProperty.SUMMARY) not in converted_dict
    assert converter.json_property_name(PackageProperty.DESCRIPTION) not in converted_dict
    assert converter.json_property_name(PackageProperty.COMMENT) not in converted_dict
    assert converter.json_property_name(PackageProperty.PRIMARY_PACKAGE_PURPOSE) not in converted_dict
    assert converter.json_property_name(PackageProperty.BUILT_DATE) not in converted_dict
    assert converter.json_property_name(PackageProperty.RELEASE_DATE) not in converted_dict
    assert converter.json_property_name(PackageProperty.VALID_UNTIL_DATE) not in converted_dict
    assert converter.json_property_name(PackageProperty.ANNOTATIONS) not in converted_dict
    assert converter.json_property_name(PackageProperty.ATTRIBUTION_TEXTS) not in converted_dict
    assert converter.json_property_name(PackageProperty.CHECKSUMS) not in converted_dict
    assert converter.json_property_name(PackageProperty.EXTERNAL_REFS) not in converted_dict
    assert converter.json_property_name(PackageProperty.LICENSE_INFO_FROM_FILES) not in converted_dict


def test_spdx_no_assertion(converter: PackageConverter):
    package = package_fixture(
        download_location=SpdxNoAssertion(),
        supplier=SpdxNoAssertion(),
        originator=SpdxNoAssertion(),
        homepage=SpdxNoAssertion(),
        license_concluded=SpdxNoAssertion(),
        license_info_from_files=[SpdxNoAssertion()],
        license_declared=SpdxNoAssertion(),
        copyright_text=SpdxNoAssertion(),
    )

    document = Document(creation_info_fixture(), packages=[package])

    converted_dict = converter.convert(package, document)

    assert converted_dict[converter.json_property_name(PackageProperty.DOWNLOAD_LOCATION)] == SPDX_NO_ASSERTION_STRING
    assert converted_dict[converter.json_property_name(PackageProperty.SUPPLIER)] == SPDX_NO_ASSERTION_STRING
    assert converted_dict[converter.json_property_name(PackageProperty.ORIGINATOR)] == SPDX_NO_ASSERTION_STRING
    assert converted_dict[converter.json_property_name(PackageProperty.HOMEPAGE)] == SPDX_NO_ASSERTION_STRING
    assert converted_dict[converter.json_property_name(PackageProperty.LICENSE_CONCLUDED)] == SPDX_NO_ASSERTION_STRING
    assert converted_dict[converter.json_property_name(PackageProperty.LICENSE_INFO_FROM_FILES)] == [
        SPDX_NO_ASSERTION_STRING
    ]
    assert converted_dict[converter.json_property_name(PackageProperty.LICENSE_DECLARED)] == SPDX_NO_ASSERTION_STRING
    assert converted_dict[converter.json_property_name(PackageProperty.COPYRIGHT_TEXT)] == SPDX_NO_ASSERTION_STRING


def test_spdx_none(converter: PackageConverter):
    package = package_fixture(
        download_location=SpdxNone(),
        homepage=SpdxNone(),
        license_concluded=SpdxNone(),
        license_info_from_files=[SpdxNone()],
        license_declared=SpdxNone(),
        copyright_text=SpdxNone(),
    )

    document = Document(creation_info_fixture(), packages=[package])

    converted_dict = converter.convert(package, document)

    assert converted_dict[converter.json_property_name(PackageProperty.DOWNLOAD_LOCATION)] == SPDX_NONE_STRING
    assert converted_dict[converter.json_property_name(PackageProperty.HOMEPAGE)] == SPDX_NONE_STRING
    assert converted_dict[converter.json_property_name(PackageProperty.LICENSE_CONCLUDED)] == SPDX_NONE_STRING
    assert converted_dict[converter.json_property_name(PackageProperty.LICENSE_INFO_FROM_FILES)] == [SPDX_NONE_STRING]
    assert converted_dict[converter.json_property_name(PackageProperty.LICENSE_DECLARED)] == SPDX_NONE_STRING
    assert converted_dict[converter.json_property_name(PackageProperty.COPYRIGHT_TEXT)] == SPDX_NONE_STRING


def test_package_annotations(converter: PackageConverter):
    package = package_fixture(spdx_id="packageId")
    document = document_fixture(packages=[package])
    first_package_annotation = annotation_fixture(spdx_id=package.spdx_id)
    second_package_annotation = annotation_fixture(spdx_id=package.spdx_id)
    document_annotation = annotation_fixture(spdx_id=document.creation_info.spdx_id)
    file_annotation = annotation_fixture(spdx_id=document.files[0].spdx_id)
    snippet_annotation = annotation_fixture(spdx_id=document.snippets[0].spdx_id)
    other_annotation = annotation_fixture(spdx_id="otherId")
    annotations = [
        first_package_annotation,
        second_package_annotation,
        document_annotation,
        file_annotation,
        snippet_annotation,
        other_annotation,
    ]
    document.annotations = annotations

    # Weird type hint to make warnings about unresolved references from the mock class disappear
    annotation_converter: Union[AnnotationConverter, NonCallableMagicMock] = converter.annotation_converter
    annotation_converter.convert.return_value = "mock_converted_annotation"

    converted_dict = converter.convert(package, document)

    assert_mock_method_called_with_arguments(
        annotation_converter, "convert", first_package_annotation, second_package_annotation
    )
    converted_file_annotations = converted_dict.get(converter.json_property_name(PackageProperty.ANNOTATIONS))
    assert converted_file_annotations == ["mock_converted_annotation", "mock_converted_annotation"]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field
from datetime import datetime
from enum import Enum, auto

from beartype.typing import Dict, List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model import CreationInfo, ExternalIdentifier, ExternalReference, IntegrityMethod
from spdx_tools.spdx3.model.licensing import LicenseField
from spdx_tools.spdx3.model.software import Package, SoftwarePurpose


class SafetyRiskAssessmentType(Enum):
    SERIOUS = auto()
    HIGH = auto()
    MEDIUM = auto()
    LOW = auto()


@dataclass_with_properties
class AIPackage(Package):
    energy_consumption: Optional[str] = None
    standard_compliance: List[str] = field(default_factory=list)
    limitation: Optional[str] = None
    type_of_model: List[str] = field(default_factory=list)
    information_about_training: Optional[str] = None
    information_about_application: Optional[str] = None
    hyperparameter: Dict[str, Optional[str]] = field(default_factory=dict)
    model_data_preprocessing: List[str] = field(default_factory=list)
    model_explainability: List[str] = field(default_factory=list)
    sensitive_personal_information: Optional[bool] = None
    metric_decision_threshold: Dict[str, Optional[str]] = field(default_factory=dict)
    metric: Dict[str, Optional[str]] = field(default_factory=dict)
    domain: List[str] = field(default_factory=list)
    autonomy_type: Optional[bool] = None
    safety_risk_assessment: Optional[SafetyRiskAssessmentType] = None

    def __init__(
        self,
        spdx_id: str,
        name: str,
        supplied_by: List[str],
        download_location: str,
        package_version: str,
        primary_purpose: SoftwarePurpose,
        release_time: datetime,
        creation_info: Optional[CreationInfo] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        comment: Optional[str] = None,
        verified_using: List[IntegrityMethod] = None,
        external_reference: List[ExternalReference] = None,
        external_identifier: List[ExternalIdentifier] = None,
        extension: Optional[str] = None,
        originated_by: List[str] = None,
        built_time: Optional[datetime] = None,
        valid_until_time: Optional[datetime] = None,
        standard: List[str] = None,
        content_identifier: Optional[str] = None,
        additional_purpose: List[SoftwarePurpose] = None,
        concluded_license: Optional[LicenseField] = None,
        declared_license: Optional[LicenseField] = None,
        copyright_text: Optional[str] = None,
        attribution_text: Optional[str] = None,
        package_url: Optional[str] = None,
        homepage: Optional[str] = None,
        source_info: Optional[str] = None,
        energy_consumption: Optional[str] = None,
        standard_compliance: List[str] = None,
        limitation: Optional[str] = None,
        type_of_model: List[str] = None,
        information_about_training: Optional[str] = None,
        information_about_application: Optional[str] = None,
        hyperparameter: Dict[str, Optional[str]] = None,
        model_data_preprocessing: List[str] = None,
        model_explainability: List[str] = None,
        sensitive_personal_information: Optional[bool] = None,
        metric_decision_threshold: Dict[str, Optional[str]] = None,
        metric: Dict[str, Optional[str]] = None,
        domain: List[str] = None,
        autonomy_type: Optional[bool] = None,
        safety_risk_assessment: Optional[SafetyRiskAssessmentType] = None,
    ):
        verified_using = [] if verified_using is None else verified_using
        external_reference = [] if external_reference is None else external_reference
        external_identifier = [] if external_identifier is None else external_identifier
        originated_by = [] if originated_by is None else originated_by
        additional_purpose = [] if additional_purpose is None else additional_purpose
        standard = [] if standard is None else standard
        standard_compliance = [] if standard_compliance is None else standard_compliance
        type_of_model = [] if type_of_model is None else type_of_model
        hyperparameter = {} if hyperparameter is None else hyperparameter
        model_data_preprocessing = [] if model_data_preprocessing is None else model_data_preprocessing
        model_explainability = [] if model_explainability is None else model_explainability
        metric_decision_threshold = {} if metric_decision_threshold is None else metric_decision_threshold
        metric = {} if metric is None else metric
        domain = [] if domain is None else domain
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest
from rdflib import RDF, RDFS, Graph, Literal, URIRef

from spdx_tools.spdx.rdfschema.namespace import LICENSE_NAMESPACE, POINTER_NAMESPACE, SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.snippet_writer import add_range_to_graph, add_snippet_to_graph
from tests.spdx.fixtures import snippet_fixture


def test_add_snippet_to_graph():
    graph = Graph()
    snippet = snippet_fixture()

    add_snippet_to_graph(snippet, graph, "docNamespace", {})

    assert (URIRef("docNamespace#SPDXRef-Snippet"), RDF.type, SPDX_NAMESPACE.Snippet) in graph
    assert (None, SPDX_NAMESPACE.snippetFromFile, URIRef(f"docNamespace#{snippet.file_spdx_id}")) in graph
    assert (None, SPDX_NAMESPACE.licenseConcluded, None) in graph
    assert (None, SPDX_NAMESPACE.licenseInfoInSnippet, LICENSE_NAMESPACE.MIT) in graph
    assert (None, SPDX_NAMESPACE.licenseInfoInSnippet, LICENSE_NAMESPACE["GPL-2.0-only"]) in graph
    assert (None, SPDX_NAMESPACE.licenseInfoInSnippet, SPDX_NAMESPACE.none) in graph
    assert (None, SPDX_NAMESPACE.licenseComments, Literal(snippet.license_comment)) in graph
    assert (None, SPDX_NAMESPACE.copyrightText, Literal(snippet.copyright_text)) in graph
    assert (None, SPDX_NAMESPACE.name, Literal(snippet.name)) in graph
    assert (None, SPDX_NAMESPACE.attributionText, Literal(snippet.attribution_texts[0])) in graph
    assert (None, RDFS.comment, Literal(snippet.comment)) in graph


@pytest.mark.parametrize(
    "range,pointer,predicate",
    [
        ((5, 190), POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.offset),
        ((1, 3), POINTER_NAMESPACE.LineCharPointer, POINTER_NAMESPACE.lineNumber),
    ],
)
def test_add_ranges_to_graph(range, pointer, predicate):
    graph = Graph()
    add_range_to_graph(range, graph, URIRef("snippetNode"), URIRef("docNamespace#SPDXRef-File"), pointer)

    assert (URIRef("snippetNode"), SPDX_NAMESPACE.range, None) in graph
    assert (None, POINTER_NAMESPACE.startPointer, None) in graph
    assert (None, POINTER_NAMESPACE.endPointer, None) in graph
    assert (None, POINTER_NAMESPACE.reference, URIRef("docNamespace#SPDXRef-File")) in graph
    assert (None, predicate, Literal(range[0])) in graph
    assert (None, predicate, Literal(range[1])) in graph



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import re

from beartype.typing import List, Optional

from spdx_tools.spdx.model import ExtractedLicensingInfo
from spdx_tools.spdx.validation.uri_validators import validate_url
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_extracted_licensing_infos(
    extracted_licensing_infos: Optional[List[ExtractedLicensingInfo]],
) -> List[ValidationMessage]:
    validation_messages = []
    for extracted_licensing_info in extracted_licensing_infos:
        validation_messages.extend(validate_extracted_licensing_info(extracted_licensing_info))

    return validation_messages


def validate_extracted_licensing_info(extracted_licensing_infos: ExtractedLicensingInfo) -> List[ValidationMessage]:
    validation_messages: List[ValidationMessage] = []
    context = ValidationContext(
        element_type=SpdxElementType.EXTRACTED_LICENSING_INFO, full_element=extracted_licensing_infos
    )

    license_id: str = extracted_licensing_infos.license_id
    if license_id and not re.match(r"^LicenseRef-[\da-zA-Z.-]+$", license_id):
        validation_messages.append(
            ValidationMessage(
                f'license_id must only contain letters, numbers, "." and "-" and must begin with "LicenseRef-", '
                f"but is: {license_id}",
                context,
            )
        )

    if license_id and not extracted_licensing_infos.extracted_text:
        validation_messages.append(
            ValidationMessage("extracted_text must be provided if there is a license_id assigned", context)
        )

    for cross_reference in extracted_licensing_infos.cross_references:
        for message in validate_url(cross_reference):
            validation_messages.append(ValidationMessage("cross_reference " + message, context))

    return validation_messages



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict, List, Optional

from spdx_tools.common.typing.constructor_type_errors import ConstructorTypeErrors
from spdx_tools.spdx.model import Relationship, RelationshipType
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import (
    delete_duplicates_from_list,
    json_str_to_enum_name,
    parse_field_or_log_error,
    parse_field_or_no_assertion_or_none,
)
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)


class RelationshipParser:
    logger: Logger

    def __init__(self):
        self.logger = Logger()

    def parse_all_relationships(self, input_doc_dict: Dict) -> List[Relationship]:
        relationships = []
        relationship_dicts: List[Dict] = input_doc_dict.get("relationships", [])
        relationships.extend(
            parse_field_or_log_error(self.logger, relationship_dicts, self.parse_relationship, [], True)
        )

        document_describes: List[str] = delete_duplicates_from_list(input_doc_dict.get("documentDescribes", []))
        doc_spdx_id: Optional[str] = input_doc_dict.get("SPDXID")

        existing_relationships_without_comments: List[Relationship] = self.get_all_relationships_without_comments(
            relationships
        )
        relationships.extend(
            parse_field_or_log_error(
                self.logger,
                document_describes,
                lambda x: self.parse_document_describes(
                    doc_spdx_id=doc_spdx_id,
                    described_spdx_ids=x,
                    existing_relationships=existing_relationships_without_comments,
                ),
                [],
            )
        )

        package_dicts: List[Dict] = input_doc_dict.get("packages", [])
        existing_relationships_without_comments: List[Relationship] = self.get_all_relationships_without_comments(
            relationships
        )

        relationships.extend(
            parse_field_or_log_error(
                self.logger,
                package_dicts,
                lambda x: self.parse_has_files(
                    package_dicts=x, existing_relationships=existing_relationships_without_comments
                ),
                [],
            )
        )

        file_dicts: List[Dict] = input_doc_dict.get("files", [])

        # not implemented yet: deal with deprecated fields in file:
        # https://github.com/spdx/tools-python/issues/294 & https://github.com/spdx/tools-python/issues/387
        _ = self.parse_artifact_of(file_dicts=file_dicts)
        _ = self.parse_file_dependencies(file_dicts=file_dicts)

        raise_parsing_error_if_logger_has_messages(self.logger)

        return relationships

    def parse_relationship(self, relationship_dict: Dict) -> Relationship:
        logger = Logger()
        spdx_element_id: Optional[str] = relationship_dict.get("spdxElementId")
        related_spdx_element: Optional[str] = parse_field_or_no_assertion_or_none(
            relationship_dict.get("relatedSpdxElement")
        )
        relationship_type: Optional[RelationshipType] = parse_field_or_log_error(
            logger, relationship_dict.get("relationshipType"), self.parse_relationship_type
        )
        relationship_comment: Optional[str] = relationship_dict.get("comment")
        raise_parsing_error_if_logger_has_messages(logger, "Relationship")

        relationship = construct_or_raise_parsing_error(
            Relationship,
            dict(
                spdx_element_id=spdx_element_id,
                relationship_type=relationship_type,
                related_spdx_element_id=related_spdx_element,
                comment=relationship_comment,
            ),
        )
        return relationship

    @staticmethod
    def parse_relationship_type(relationship_type_str: str) -> RelationshipType:
        try:
            relationship_type = RelationshipType[json_str_to_enum_name(relationship_type_str)]
        except KeyError:
            raise SPDXParsingError([f"Invalid RelationshipType: {relationship_type_str}"])
        return relationship_type

    def parse_document_describes(
        self, doc_spdx_id: str, described_spdx_ids: List[str], existing_relationships: List[Relationship]
    ) -> List[Relationship]:
        logger = Logger()
        describes_relationships = []
        for spdx_id in described_spdx_ids:
            try:
                describes_relationship = Relationship(
                    spdx_element_id=doc_spdx_id,
                    relationship_type=RelationshipType.DESCRIBES,
                    related_spdx_element_id=spdx_id,
                )
            except ConstructorTypeErrors as err:
                logger.append(err.get_messages())
                continue
            if not self.check_if_relationship_exists(describes_relationship, existing_relationships):
                describes_relationships.append(describes_relationship)
        raise_parsing_error_if_logger_has_messages(logger, "document describes relationships")

        return describes_relationships

    def parse_has_files(
        self, package_dicts: List[Dict], existing_relationships: List[Relationship]
    ) -> List[Relationship]:
        # assume existing relationships are stripped of comments
        logger = Logger()
        contains_relationships = []
        for package in package_dicts:
            package_spdx_id: Optional[str] = package.get("SPDXID")
            contained_files: List[str] = delete_duplicates_from_list(package.get("hasFiles", []))
            if not contained_files:
                continue
            for file_spdx_id in contained_files:
                try:
                    contains_relationship = Relationship(
                        spdx_element_id=package_spdx_id,
                        relationship_type=RelationshipType.CONTAINS,
                        related_spdx_element_id=file_spdx_id,
                    )
                except ConstructorTypeErrors as err:
                    logger.append(err.get_messages())
                    continue
                if not self.check_if_relationship_exists(
                    relationship=contains_relationship, existing_relationships=existing_relationships
                ):
                    contains_relationships.append(contains_relationship)
        raise_parsing_error_if_logger_has_messages(logger, "package contains relationships")

        return contains_relationships

    def check_if_relationship_exists(
        self, relationship: Relationship, existing_relationships: List[Relationship]
    ) -> bool:
        # assume existing relationships are stripped of comments
        if relationship in existing_relationships:
            return True
        relationship_inverted: Relationship = self.invert_relationship(relationship)
        if relationship_inverted in existing_relationships:
            return True

        return False

    @staticmethod
    def get_all_relationships_without_comments(existing_relationships: List[Relationship]) -> List[Relationship]:
        relationships_without_comments = [
            Relationship(
                relationship_type=relationship.relationship_type,
                related_spdx_element_id=relationship.related_spdx_element_id,
                spdx_element_id=relationship.spdx_element_id,
            )
            for relationship in existing_relationships
        ]
        return relationships_without_comments

    def invert_relationship(self, relationship: Relationship) -> Relationship:
        return Relationship(
            related_spdx_element_id=relationship.spdx_element_id,
            spdx_element_id=relationship.related_spdx_element_id,
            relationship_type=self.invert_relationship_types[relationship.relationship_type],
            comment=relationship.comment,
        )

    invert_relationship_types = {
        RelationshipType.DESCRIBES: RelationshipType.DESCRIBED_BY,
        RelationshipType.DESCRIBED_BY: RelationshipType.DESCRIBES,
        RelationshipType.CONTAINS: RelationshipType.CONTAINED_BY,
        RelationshipType.CONTAINED_BY: RelationshipType.CONTAINS,
    }

    @staticmethod
    def parse_file_dependencies(file_dicts: List[Dict]) -> List[Relationship]:
        dependency_relationships = []
        # the field fileDependencies is deprecated and should be converted to a relationship
        # https://github.com/spdx/tools-python/issues/387
        return dependency_relationships

    @staticmethod
    def parse_artifact_of(file_dicts: List[Dict]) -> List[Relationship]:
        generated_relationships = []
        # artifactOfs is deprecated and should be converted to an external package and a generated from relationship
        # https://github.com/spdx/tools-python/issues/294
        return generated_relationships



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Optional, Union
from license_expression import LicenseExpression
from rdflib import RDF, Graph
from rdflib.term import BNode, Identifier, Node, URIRef

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.rdf.graph_parsing_functions import get_value_from_graph, remove_prefix
from spdx_tools.spdx.rdfschema.namespace import LICENSE_NAMESPACE, SPDX_NAMESPACE


def parse_license_expression(
    license_expression_node: Union[URIRef, BNode, Node],
    graph: Graph,
    doc_namespace: str,
    logger: Optional[Logger] = None,
) -> LicenseExpression:
    if not logger:
        logger = Logger()

    expression = ""
    if license_expression_node.startswith(LICENSE_NAMESPACE):
        expression = remove_prefix(license_expression_node, LICENSE_NAMESPACE)
        return spdx_licensing.parse(expression)
    if license_expression_node.startswith(doc_namespace):
        expression = license_expression_node.fragment
        return spdx_licensing.parse(expression)

    node_type = graph.value(license_expression_node, RDF.type)
    if node_type == SPDX_NAMESPACE.ConjunctiveLicenseSet:
        members = []
        for _, _, member_node in graph.triples((license_expression_node, SPDX_NAMESPACE.member, None)):
            members.append(parse_license_expression(member_node, graph, doc_namespace, logger))
        expression = " AND ".join([str(member) for member in members])
    if node_type == SPDX_NAMESPACE.DisjunctiveLicenseSet:
        members = []
        for _, _, member_node in graph.triples((license_expression_node, SPDX_NAMESPACE.member, None)):
            members.append(parse_license_expression(member_node, graph, doc_namespace, logger))
        expression = " OR ".join([str(member) for member in members])
    if node_type == SPDX_NAMESPACE.WithExceptionOperator:
        license_expression = parse_license_expression(
            graph.value(license_expression_node, SPDX_NAMESPACE.member), graph, doc_namespace, logger
        )
        exception = parse_license_exception(
            get_value_from_graph(logger, graph, license_expression_node, SPDX_NAMESPACE.licenseException),
            graph,
            logger,
        )
        expression = f"{license_expression} WITH {exception}"

    return spdx_licensing.parse(expression)


def parse_license_exception(exception_node: Identifier, graph: Graph, logger) -> str:
    if exception_node.startswith(LICENSE_NAMESPACE):
        exception = remove_prefix(exception_node, LICENSE_NAMESPACE)
    else:
        exception = get_value_from_graph(logger, graph, exception_node, SPDX_NAMESPACE.licenseExceptionId).toPython()
    return exception



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock, call, mock_open, patch

from spdx_tools.spdx.writer.tagvalue.extracted_licensing_info_writer import write_extracted_licensing_info
from tests.spdx.fixtures import extracted_licensing_info_fixture


def test_extracted_licensing_info_writer():
    extracted_licensing_info = extracted_licensing_info_fixture()

    mock: MagicMock = mock_open()
    with patch(f"{__name__}.open", mock, create=True):
        with open("foo", "w") as file:
            write_extracted_licensing_info(extracted_licensing_info, file)

    mock.assert_called_once_with("foo", "w")
    handle = mock()
    handle.write.assert_has_calls(
        [
            call(f"LicenseID: {extracted_licensing_info.license_id}\n"),
            call(f"ExtractedText: {extracted_licensing_info.extracted_text}\n"),
            call(f"LicenseName: {extracted_licensing_info.license_name}\n"),
            call(f"LicenseCrossReference: {extracted_licensing_info.cross_references[0]}\n"),
            call(f"LicenseComment: {extracted_licensing_info.comment}\n"),
        ]
    )



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from enum import Enum, auto

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx.model import Actor


class AnnotationType(Enum):
    REVIEW = auto()
    OTHER = auto()


@dataclass_with_properties
class Annotation:
    spdx_id: str
    annotation_type: AnnotationType
    annotator: Actor
    annotation_date: datetime
    annotation_comment: str

    def __init__(
        self,
        spdx_id: str,
        annotation_type: AnnotationType,
        annotator: Actor,
        annotation_date: datetime,
        annotation_comment: str,
    ):
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os
from unittest import TestCase

import pytest
from rdflib import RDF, BNode, Graph, Literal, URIRef

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.model import SpdxNoAssertion
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.rdf.snippet_parser import parse_ranges, parse_snippet
from spdx_tools.spdx.rdfschema.namespace import POINTER_NAMESPACE, SPDX_NAMESPACE


def test_parse_snippet():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    snippet_node = graph.value(predicate=RDF.type, object=SPDX_NAMESPACE.Snippet)
    doc_namespace = "https://some.namespace"
    assert isinstance(snippet_node, URIRef)

    snippet = parse_snippet(snippet_node, graph, doc_namespace)

    assert snippet.spdx_id == "SPDXRef-Snippet"
    assert snippet.file_spdx_id == "SPDXRef-File"
    assert snippet.byte_range == (1, 2)
    assert snippet.line_range == (3, 4)
    assert snippet.license_concluded == spdx_licensing.parse("MIT AND GPL-2.0")
    TestCase().assertCountEqual(
        snippet.license_info_in_snippet,
        [spdx_licensing.parse("MIT"), spdx_licensing.parse("GPL-2.0"), SpdxNoAssertion()],
    )
    assert snippet.license_comment == "snippetLicenseComment"
    assert snippet.copyright_text == "licenseCopyrightText"
    assert snippet.comment == "snippetComment"
    assert snippet.name == "snippetName"
    assert snippet.attribution_texts == ["snippetAttributionText"]


@pytest.mark.parametrize(
    "predicate_value_class_member",
    [
        (
            [
                (POINTER_NAMESPACE.startPointer, 1, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.offset),
                (POINTER_NAMESPACE.endPointer, 2, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.offset),
            ]
        ),
        (
            [
                (POINTER_NAMESPACE.startPointer, 100, POINTER_NAMESPACE.LineCharPointer, POINTER_NAMESPACE.lineNumber),
                (POINTER_NAMESPACE.endPointer, 200, POINTER_NAMESPACE.LineCharPointer, POINTER_NAMESPACE.lineNumber),
            ]
        ),
    ],
)
def test_parse_ranges(predicate_value_class_member):
    graph = Graph()
    pointer_class = predicate_value_class_member[0][2]

    add_range_to_graph_helper(graph, predicate_value_class_member)

    range_node = graph.value(predicate=RDF.type, object=POINTER_NAMESPACE.StartEndPointer)
    assert isinstance(range_node, BNode)
    range_dict = parse_ranges(range_node, graph)

    assert pointer_class.fragment in range_dict.keys()
    assert range_dict[pointer_class.fragment][0] == predicate_value_class_member[0][1]
    assert range_dict[pointer_class.fragment][1] == predicate_value_class_member[1][1]


@pytest.mark.parametrize(
    "predicate_value_class_member",
    [
        (
            [
                (POINTER_NAMESPACE.startPointer, 1, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.lineNumber),
                (POINTER_NAMESPACE.endPointer, 2, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.lineNumber),
            ]
        ),
        (
            [
                (POINTER_NAMESPACE.startPointer, 100, POINTER_NAMESPACE.LineCharPointer, POINTER_NAMESPACE.offset),
                (POINTER_NAMESPACE.endPointer, 200, POINTER_NAMESPACE.LineCharPointer, POINTER_NAMESPACE.offset),
            ]
        ),
    ],
)
def test_parse_ranges_wrong_pair_of_pointer_classes(predicate_value_class_member):
    graph = Graph()
    pointer_class = predicate_value_class_member[0][2]

    add_range_to_graph_helper(graph, predicate_value_class_member)

    range_node = graph.value(predicate=RDF.type, object=POINTER_NAMESPACE.StartEndPointer)
    assert isinstance(range_node, BNode)
    range_dict = parse_ranges(range_node, graph)

    assert pointer_class.fragment in range_dict.keys()
    assert range_dict[pointer_class.fragment][0] is None
    assert range_dict[pointer_class.fragment][1] is None


@pytest.mark.parametrize(
    "predicate_value_class_member,expected_message",
    [
        (
            [
                (POINTER_NAMESPACE.endPointer, 1, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.offset),
                (POINTER_NAMESPACE.endPointer, 2, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.offset),
            ],
            "Couldn't find pointer of type startPointer.",
        ),
        (
            [(POINTER_NAMESPACE.startPointer, 1, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.offset)],
            "Couldn't find pointer of type endPointer.",
        ),
        (
            [
                (POINTER_NAMESPACE.startPointer, 1, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.offset),
                (POINTER_NAMESPACE.endPointer, 2, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.offset),
                (POINTER_NAMESPACE.endPointer, 3, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.offset),
            ],
            "Multiple values for endPointer.",
        ),
        (
            [
                (POINTER_NAMESPACE.startPointer, 100, POINTER_NAMESPACE.LineCharPointer, POINTER_NAMESPACE.lineNumber),
                (POINTER_NAMESPACE.startPointer, 200, POINTER_NAMESPACE.LineCharPointer, POINTER_NAMESPACE.lineNumber),
            ],
            "Multiple values for startPointer",
        ),
        (
            [
                (POINTER_NAMESPACE.startPointer, 100, POINTER_NAMESPACE.LineCharPointer, POINTER_NAMESPACE.lineNumber),
                (POINTER_NAMESPACE.endPointer, 200, POINTER_NAMESPACE.ByteOffsetPointer, POINTER_NAMESPACE.offset),
            ],
            "Types of startPointer and endPointer don't match",
        ),
    ],
)
def test_parse_ranges_error(predicate_value_class_member, expected_message):
    graph = Graph()

    add_range_to_graph_helper(graph, predicate_value_class_member)

    with pytest.raises(SPDXParsingError, match=expected_message):
        range_node = graph.value(predicate=RDF.type, object=POINTER_NAMESPACE.StartEndPointer)
        assert isinstance(range_node, BNode)
        parse_ranges(range_node, graph)


def add_range_to_graph_helper(graph, predicate_value_class_member):
    start_end_pointer = BNode()
    graph.add((start_end_pointer, RDF.type, POINTER_NAMESPACE.StartEndPointer))
    for predicate, value, pointer_class, pointer_member in predicate_value_class_member:
        pointer_node = BNode()
        graph.add((pointer_node, RDF.type, pointer_class))
        graph.add((start_end_pointer, predicate, pointer_node))
        graph.add((pointer_node, pointer_member, Literal(value)))


def test_parse_invalid_file():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/invalid_documents/file_without_spdx_ids.xml"))
    snippet_node = graph.value(predicate=RDF.type, object=SPDX_NAMESPACE.Snippet)
    doc_namespace = "https://some.namespace"

    assert isinstance(snippet_node, BNode)
    with pytest.raises(SPDXParsingError):
        parse_snippet(snippet_node, graph, doc_namespace)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List

import pytest

from spdx_tools.spdx.model import Annotation, Document
from spdx_tools.spdx.validation.annotation_validator import validate_annotation
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import annotation_fixture, document_fixture, file_fixture


def test_valid_annotation():
    validation_messages: List[ValidationMessage] = validate_annotation(annotation_fixture(), document_fixture())

    assert validation_messages == []


@pytest.mark.parametrize(
    "annotation_id, file_id, expected_message",
    [
        (
            "SPDXRef-File",
            "SPDXRef-hiddenFile",
            'did not find the referenced spdx_id "SPDXRef-File" in the SPDX document',
        )
    ],
)
def test_invalid_annotation(annotation_id, file_id, expected_message):
    annotation: Annotation = annotation_fixture(spdx_id=annotation_id)
    document: Document = document_fixture(files=[file_fixture(spdx_id=file_id)])
    validation_messages: List[ValidationMessage] = validate_annotation(annotation, document)

    expected = ValidationMessage(
        expected_message, ValidationContext(element_type=SpdxElementType.ANNOTATION, full_element=annotation)
    )

    assert validation_messages == [expected]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import xmltodict
from beartype.typing import Any, Dict

from spdx_tools.spdx.model import Document
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.json_like_dict_parser import JsonLikeDictParser

LIST_LIKE_FIELDS = [
    "creators",
    "externalDocumentRefs",
    "hasExtractedLicensingInfos",
    "seeAlsos",
    "annotations",
    "relationships",
    "snippets",
    "reviewers",
    "fileTypes",
    "licenseInfoFromFiles",
    "licenseInfoInFiles",
    "artifactOf",
    "fileContributors",
    "fileDependencies",
    "files",
    "documentDescribes",
    "packages",
    "checksums",
    "hasFiles",
    "externalRefs",
    "ranges",
    "licenseInfoInSnippets",
    "packageVerificationCodeExcludedFiles",
    "attributionTexts",
]


def parse_from_file(file_name: str, encoding: str = "utf-8") -> Document:
    with open(file_name, encoding=encoding) as file:
        parsed_xml: Dict = xmltodict.parse(file.read(), encoding="utf-8")

    input_doc_as_dict: Dict = _fix_list_like_fields(parsed_xml).get("Document")

    if not input_doc_as_dict:
        raise SPDXParsingError(['Did not find the XML top level tag "Document".'])

    return JsonLikeDictParser().parse(input_doc_as_dict)


def _fix_list_like_fields(data: Any) -> Any:
    """
    XML files do not contain lists. Thus, single fields that should be a list in SPDX have to be manually cast.
    This method takes a parsed dictionary and converts all values with key from LIST_LIKE_FIELDS to lists.
    """
    if isinstance(data, dict):
        new_data = {}
        for key, value in data.items():
            if key in LIST_LIKE_FIELDS and not isinstance(value, list):
                new_data[key] = [_fix_list_like_fields(value)] if value else []
            else:
                new_data[key] = _fix_list_like_fields(value)
        return new_data

    if isinstance(data, list):
        new_data = []
        for element in data:
            new_data.append(_fix_list_like_fields(element))
        return new_data

    return data



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Type

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.jsonschema.annotation_converter import AnnotationConverter
from spdx_tools.spdx.jsonschema.checksum_converter import ChecksumConverter
from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.external_package_ref_converter import ExternalPackageRefConverter
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.jsonschema.optional_utils import apply_if_present
from spdx_tools.spdx.jsonschema.package_properties import PackageProperty
from spdx_tools.spdx.jsonschema.package_verification_code_converter import PackageVerificationCodeConverter
from spdx_tools.spdx.model import Actor, Document, Package


class PackageConverter(TypedConverter[Package]):
    annotation_converter: AnnotationConverter
    checksum_converter: ChecksumConverter
    external_package_ref_converter: ExternalPackageRefConverter
    package_verification_code_converter: PackageVerificationCodeConverter

    def __init__(self):
        self.annotation_converter = AnnotationConverter()
        self.checksum_converter = ChecksumConverter()
        self.external_package_ref_converter = ExternalPackageRefConverter()
        self.package_verification_code_converter = PackageVerificationCodeConverter()

    def json_property_name(self, package_property: PackageProperty) -> str:
        if package_property == PackageProperty.SPDX_ID:
            return "SPDXID"
        return super().json_property_name(package_property)

    def _get_property_value(
        self, package: Package, package_property: PackageProperty, document: Document = None
    ) -> Any:
        if package_property == PackageProperty.SPDX_ID:
            return package.spdx_id
        elif package_property == PackageProperty.ANNOTATIONS:
            package_annotations = filter(
                lambda annotation: annotation.spdx_id == package.spdx_id, document.annotations
            )
            return [
                self.annotation_converter.convert(annotation, document) for annotation in package_annotations
            ] or None
        elif package_property == PackageProperty.ATTRIBUTION_TEXTS:
            return package.attribution_texts or None
        elif package_property == PackageProperty.BUILT_DATE:
            return apply_if_present(datetime_to_iso_string, package.built_date)
        elif package_property == PackageProperty.CHECKSUMS:
            return [self.checksum_converter.convert(checksum, document) for checksum in package.checksums] or None
        elif package_property == PackageProperty.COMMENT:
            return package.comment
        elif package_property == PackageProperty.COPYRIGHT_TEXT:
            return apply_if_present(str, package.copyright_text)
        elif package_property == PackageProperty.DESCRIPTION:
            return package.description
        elif package_property == PackageProperty.DOWNLOAD_LOCATION:
            return str(package.download_location)
        elif package_property == PackageProperty.EXTERNAL_REFS:
            return [
                self.external_package_ref_converter.convert(external_ref)
                for external_ref in package.external_references
            ] or None
        elif package_property == PackageProperty.FILES_ANALYZED:
            return package.files_analyzed
        elif package_property == PackageProperty.HOMEPAGE:
            return apply_if_present(str, package.homepage)
        elif package_property == PackageProperty.LICENSE_COMMENTS:
            return package.license_comment
        elif package_property == PackageProperty.LICENSE_CONCLUDED:
            return apply_if_present(str, package.license_concluded)
        elif package_property == PackageProperty.LICENSE_DECLARED:
            return apply_if_present(str, package.license_declared)
        elif package_property == PackageProperty.LICENSE_INFO_FROM_FILES:
            return [str(license_expression) for license_expression in package.license_info_from_files] or None
        elif package_property == PackageProperty.NAME:
            return package.name
        elif package_property == PackageProperty.ORIGINATOR:
            if isinstance(package.originator, Actor):
                return package.originator.to_serialized_string()
            return apply_if_present(str, package.originator)
        elif package_property == PackageProperty.PACKAGE_FILE_NAME:
            return package.file_name
        elif package_property == PackageProperty.PACKAGE_VERIFICATION_CODE:
            return apply_if_present(self.package_verification_code_converter.convert, package.verification_code)
        elif package_property == PackageProperty.PRIMARY_PACKAGE_PURPOSE:
            return package.primary_package_purpose.name if package.primary_package_purpose is not None else None
        elif package_property == PackageProperty.RELEASE_DATE:
            return apply_if_present(datetime_to_iso_string, package.release_date)
        elif package_property == PackageProperty.SOURCE_INFO:
            return package.source_info
        elif package_property == PackageProperty.SUMMARY:
            return package.summary
        elif package_property == PackageProperty.SUPPLIER:
            if isinstance(package.supplier, Actor):
                return package.supplier.to_serialized_string()
            return apply_if_present(str, package.supplier)
        elif package_property == PackageProperty.VALID_UNTIL_DATE:
            return apply_if_present(datetime_to_iso_string, package.valid_until_date)
        elif package_property == PackageProperty.VERSION_INFO:
            return package.version

    def get_json_type(self) -> Type[JsonProperty]:
        return PackageProperty

    def get_data_model_type(self) -> Type[Package]:
        return Package

    def requires_full_document(self) -> bool:
        return True



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List
from unittest import TestCase

import pytest

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import Relationship, RelationshipType, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.validation.package_validator import validate_package, validate_package_within_document
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import document_fixture, file_fixture, package_fixture, package_verification_code_fixture


def test_valid_package():
    package = package_fixture()
    validation_messages: List[ValidationMessage] = validate_package_within_document(
        package, "SPDX-2.3", document_fixture()
    )

    assert validation_messages == []


@pytest.mark.parametrize(
    "package_input, expected_message",
    [
        (
            package_fixture(
                files_analyzed=False, verification_code=package_verification_code_fixture(), license_info_from_files=[]
            ),
            f"verification_code must be None if files_analyzed is False, but is: "
            f"{package_verification_code_fixture()}",
        ),
        (
            package_fixture(files_analyzed=False, license_info_from_files=[SpdxNone()], verification_code=None),
            "license_info_from_files must be None if files_analyzed is False, but is: [NONE]",
        ),
        (
            package_fixture(files_analyzed=False, license_info_from_files=[SpdxNoAssertion()], verification_code=None),
            "license_info_from_files must be None if files_analyzed is False, but is: [NOASSERTION]",
        ),
        (
            package_fixture(
                files_analyzed=False,
                license_info_from_files=[spdx_licensing.parse("some_license")],
                verification_code=None,
            ),
            "license_info_from_files must be None if files_analyzed is False, but is: [LicenseSymbol('some_license', "
            "is_exception=False)]",
        ),
    ],
)
def test_invalid_package(package_input, expected_message):
    validation_messages: List[ValidationMessage] = validate_package_within_document(
        package_input, "SPDX-2.3", document_fixture(relationships=[])
    )

    expected = ValidationMessage(
        expected_message,
        ValidationContext(
            spdx_id=package_input.spdx_id,
            parent_id=DOCUMENT_SPDX_ID,
            element_type=SpdxElementType.PACKAGE,
            full_element=package_input,
        ),
    )

    assert validation_messages == [expected]


@pytest.mark.parametrize(
    "relationships",
    [
        [Relationship("SPDXRef-Package", RelationshipType.CONTAINS, "DocumentRef-external:SPDXRef-File")],
        [Relationship("DocumentRef-external:SPDXRef-File", RelationshipType.CONTAINED_BY, "SPDXRef-Package")],
    ],
)
def test_valid_package_with_contains(relationships):
    document = document_fixture(
        relationships=relationships,
        files=[file_fixture(spdx_id="SPDXRef-File1"), file_fixture(spdx_id="SPDXRef-File2")],
    )
    package = package_fixture(files_analyzed=False, verification_code=None, license_info_from_files=[])

    validation_messages: List[ValidationMessage] = validate_package_within_document(package, "SPDX-2.3", document)

    assert validation_messages == []


@pytest.mark.parametrize(
    "relationships",
    [
        [Relationship("SPDXRef-Package", RelationshipType.CONTAINS, "SPDXRef-File1")],
        [Relationship("SPDXRef-File2", RelationshipType.CONTAINED_BY, "SPDXRef-Package")],
        [
            Relationship("SPDXRef-Package", RelationshipType.CONTAINS, "SPDXRef-File2"),
            Relationship("SPDXRef-File1", RelationshipType.CONTAINED_BY, "SPDXRef-Package"),
        ],
    ],
)
def test_invalid_package_with_contains(relationships):
    document = document_fixture(
        relationships=relationships,
        files=[file_fixture(spdx_id="SPDXRef-File1"), file_fixture(spdx_id="SPDXRef-File2")],
    )
    package = package_fixture(files_analyzed=False, verification_code=None, license_info_from_files=[])
    context = ValidationContext(
        spdx_id=package.spdx_id,
        parent_id=document.creation_info.spdx_id,
        element_type=SpdxElementType.PACKAGE,
        full_element=package,
    )

    validation_messages: List[ValidationMessage] = validate_package_within_document(package, "SPDX-2.3", document)

    assert validation_messages == [
        ValidationMessage(
            f"package must contain no elements if files_analyzed is False, but found {relationships}", context
        )
    ]


def test_v2_3only_fields():
    package = package_fixture()
    validation_messages: List[ValidationMessage] = validate_package(package, "SPDX-2.2")

    context = ValidationContext(spdx_id=package.spdx_id, element_type=SpdxElementType.PACKAGE, full_element=package)
    unsupported_fields = ["primary_package_purpose", "built_date", "release_date", "valid_until_date"]
    expected = [ValidationMessage(f"{field} is not supported in SPDX-2.2", context) for field in unsupported_fields]

    TestCase().assertCountEqual(validation_messages, expected)


def test_v2_2mandatory_fields():
    package = package_fixture(
        license_concluded=None,
        license_declared=None,
        copyright_text=None,
        primary_package_purpose=None,
        built_date=None,
        release_date=None,
        valid_until_date=None,
    )

    assert validate_package(package, "SPDX-2.3") == []

    validation_messages: List[ValidationMessage] = validate_package(package, "SPDX-2.2")

    context = ValidationContext(spdx_id=package.spdx_id, element_type=SpdxElementType.PACKAGE, full_element=package)
    mandatory_fields = ["license_concluded", "license_declared", "copyright_text"]
    expected = [ValidationMessage(f"{field} is mandatory in SPDX-2.2", context) for field in mandatory_fields]

    TestCase().assertCountEqual(validation_messages, expected)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field
from datetime import datetime
from enum import Enum, auto

from beartype.typing import Dict, List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model import CreationInfo, ExternalIdentifier, ExternalReference, IntegrityMethod
from spdx_tools.spdx3.model.licensing import LicenseField
from spdx_tools.spdx3.model.software import Package, SoftwarePurpose


class DatasetType(Enum):
    STRUCTURED = auto()
    NUMERIC = auto()
    TEXT = auto()
    CATEGORICAL = auto()
    GRAPH = auto()
    TIMESERIES = auto()
    TIMESTAMP = auto()
    SENSOR = auto()
    IMAGE = auto()
    SYNTACTIC = auto()
    AUDIO = auto()
    VIDEO = auto()
    OTHER = auto()
    NO_ASSERTION = auto()


class ConfidentialityLevelType(Enum):
    RED = auto()
    AMBER = auto()
    GREEN = auto()
    CLEAR = auto()


class DatasetAvailabilityType(Enum):
    DIRECT_DOWNLOAD = auto()
    SCRAPING_SCRIPT = auto()
    QUERY = auto()
    CLICKTHROUGH = auto()
    REGISTRATION = auto()


@dataclass_with_properties
class Dataset(Package):
    dataset_type: List[DatasetType] = None
    data_collection_process: Optional[str] = None
    intended_use: Optional[str] = None
    dataset_size: Optional[int] = None
    dataset_noise: Optional[str] = None
    data_preprocessing: List[str] = field(default_factory=list)
    sensor: Dict[str, Optional[str]] = field(default_factory=dict)
    known_bias: List[str] = field(default_factory=list)
    sensitive_personal_information: Optional[bool] = None
    anonymization_method_used: List[str] = field(default_factory=list)
    confidentiality_level: Optional[ConfidentialityLevelType] = None
    dataset_update_mechanism: Optional[str] = None
    dataset_availability: Optional[DatasetAvailabilityType] = None

    def __init__(
        self,
        spdx_id: str,
        name: str,
        originated_by: List[str],
        download_location: str,
        primary_purpose: SoftwarePurpose,
        built_time: datetime,
        release_time: datetime,
        dataset_type: List[DatasetType],
        creation_info: Optional[CreationInfo] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        comment: Optional[str] = None,
        verified_using: List[IntegrityMethod] = None,
        external_reference: List[ExternalReference] = None,
        external_identifier: List[ExternalIdentifier] = None,
        extension: Optional[str] = None,
        supplied_by: List[str] = None,
        valid_until_time: Optional[datetime] = None,
        standard: List[str] = None,
        content_identifier: Optional[str] = None,
        additional_purpose: List[SoftwarePurpose] = None,
        concluded_license: Optional[LicenseField] = None,
        declared_license: Optional[LicenseField] = None,
        copyright_text: Optional[str] = None,
        attribution_text: Optional[str] = None,
        package_version: Optional[str] = None,
        package_url: Optional[str] = None,
        homepage: Optional[str] = None,
        source_info: Optional[str] = None,
        data_collection_process: Optional[str] = None,
        intended_use: Optional[str] = None,
        dataset_size: Optional[int] = None,
        dataset_noise: Optional[str] = None,
        data_preprocessing: List[str] = None,
        sensor: Dict[str, Optional[str]] = None,
        known_bias: List[str] = None,
        sensitive_personal_information: Optional[bool] = None,
        anonymization_method_used: List[str] = None,
        confidentiality_level: Optional[ConfidentialityLevelType] = None,
        dataset_update_mechanism: Optional[str] = None,
        dataset_availability: Optional[DatasetAvailabilityType] = None,
    ):
        verified_using = [] if verified_using is None else verified_using
        external_reference = [] if external_reference is None else external_reference
        external_identifier = [] if external_identifier is None else external_identifier
        originated_by = [] if originated_by is None else originated_by
        additional_purpose = [] if additional_purpose is None else additional_purpose
        supplied_by = [] if supplied_by is None else supplied_by
        standard = [] if standard is None else standard
        data_preprocessing = [] if data_preprocessing is None else data_preprocessing
        sensor = {} if sensor is None else sensor
        known_bias = [] if known_bias is None else known_bias
        anonymization_method_used = [] if anonymization_method_used is None else anonymization_method_used
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.spdx.model import Document
from spdx_tools.spdx.parser.tagvalue.parser import Parser


def parse_from_file(file_name: str, encoding: str = "utf-8") -> Document:
    parser = Parser()
    with open(file_name, encoding=encoding) as file:
        data = file.read()
    document: Document = parser.parse(data)
    return document



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict

from spdx_tools.spdx3.model import Element


class Payload:
    _spdx_id_map: Dict[str, Element]

    def __init__(self, spdx_id_map: Dict[str, Element] = None):
        self._spdx_id_map = spdx_id_map if spdx_id_map else {}

    def add_element(self, element: Element):
        self._spdx_id_map[element.spdx_id] = element

    def get_element(self, spdx_id: str) -> Element:
        return self._spdx_id_map[spdx_id]

    def get_full_map(self) -> Dict[str, Element]:
        return self._spdx_id_map



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest
from rdflib import Graph, Namespace, URIRef

from spdx_tools.spdx.parser.rdf.graph_parsing_functions import parse_spdx_id, remove_prefix


@pytest.mark.parametrize(
    "resource,doc_namespace,ext_namespace_mapping,expected",
    [
        (URIRef("docNamespace#SPDXRef-Test"), "docNamespace", ("", Namespace("")), "SPDXRef-Test"),
        (URIRef("docNamespaceSPDXRef-Test"), "docNamespace", ("", Namespace("")), "docNamespaceSPDXRef-Test"),
        (
            URIRef("differentNamespace#SPDXRef-Test"),
            "docNamespace",
            ("extDoc", Namespace("differentNamespace#")),
            "extDoc:SPDXRef-Test",
        ),
        (None, "", ("", Namespace("")), None),
    ],
)
def test_parse_spdx_id(resource, doc_namespace, ext_namespace_mapping, expected):
    graph = Graph()
    graph.bind(*ext_namespace_mapping)
    spdx_id = parse_spdx_id(resource, doc_namespace, graph)

    assert spdx_id == expected


@pytest.mark.parametrize(
    "string,prefix,expected", [("prefixString", "prefix", "String"), ("prefixString", "refix", "prefixString")]
)
def test_remove_prefix(string, prefix, expected):
    shorten_string = remove_prefix(string, prefix)

    assert expected == shorten_string



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from re import sub


def snake_case_to_camel_case(snake_case_string: str) -> str:
    each_word_capitalized = sub(r"[_\-]+", " ", snake_case_string).title().replace(" ", "")
    return each_word_capitalized[0].lower() + each_word_capitalized[1:]


def camel_case_to_snake_case(camel_case_string: str) -> str:
    snake_case_string = sub("(?!^)([A-Z]+)", r"_\1", camel_case_string).lower()
    return snake_case_string



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model import (
    Bundle,
    CreationInfo,
    ExternalIdentifier,
    ExternalMap,
    ExternalReference,
    IntegrityMethod,
    NamespaceMap,
)


@dataclass_with_properties
class Bom(Bundle):
    # We overwrite the super-__init__ as check_types_and_set_values()
    # takes care of all fields (including inherited ones).

    def __init__(
        self,
        spdx_id: str,
        element: List[str],
        root_element: List[str],
        creation_info: Optional[CreationInfo] = None,
        name: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        comment: Optional[str] = None,
        verified_using: List[IntegrityMethod] = None,
        external_reference: List[ExternalReference] = None,
        external_identifier: List[ExternalIdentifier] = None,
        extension: Optional[str] = None,
        namespaces: List[NamespaceMap] = None,
        imports: List[ExternalMap] = None,
        context: Optional[str] = None,
    ):
        verified_using = [] if verified_using is None else verified_using
        external_reference = [] if external_reference is None else external_reference
        external_identifier = [] if external_identifier is None else external_identifier
        namespaces = [] if namespaces is None else namespaces
        imports = [] if imports is None else imports
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class ExternalPackageRefProperty(JsonProperty):
    COMMENT = auto()
    REFERENCE_CATEGORY = auto()
    REFERENCE_LOCATOR = auto()
    REFERENCE_TYPE = auto()



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod

from beartype.typing import Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties


@dataclass_with_properties
class IntegrityMethod(ABC):
    comment: Optional[str] = None

    @abstractmethod
    def __init__(self):
        pass



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import re

from beartype.typing import Match, Pattern

from spdx_tools.spdx.model import Actor, ActorType
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.parsing_functions import construct_or_raise_parsing_error


class ActorParser:
    @staticmethod
    def parse_actor(actor: str) -> Actor:
        tool_re: Pattern = re.compile(r"^Tool:\s*(.+)", re.UNICODE)
        person_re: Pattern = re.compile(r"^Person:\s*(?:(.*)\((.*)\)|(.*))$", re.UNICODE)
        org_re: Pattern = re.compile(r"^Organization:\s*(?:(.*)\((.*)\)|(.*))$", re.UNICODE)
        tool_match: Match = tool_re.match(actor)
        person_match: Match = person_re.match(actor)
        org_match: Match = org_re.match(actor)

        if tool_match:
            name: str = tool_match.group(1).strip()
            if not name:
                raise SPDXParsingError([f"No name for Tool provided: {actor}."])
            return construct_or_raise_parsing_error(Actor, dict(actor_type=ActorType.TOOL, name=name))

        if person_match:
            actor_type = ActorType.PERSON
            match = person_match
        elif org_match:
            actor_type = ActorType.ORGANIZATION
            match = org_match
        else:
            raise SPDXParsingError([f"Actor {actor} doesn't match any of person, organization or tool."])

        if match.group(3):
            return construct_or_raise_parsing_error(
                Actor, dict(actor_type=actor_type, name=match.group(3).strip(), email=None)
            )
        else:
            name = match.group(1)
            if not name:
                raise SPDXParsingError([f"No name for Actor provided: {actor}."])
            else:
                name = name.strip()

            email = match.group(2).strip()

            return construct_or_raise_parsing_error(
                Actor, dict(actor_type=actor_type, name=name, email=email if email else None)
            )



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List

import pytest

from spdx_tools.spdx.model import ExternalPackageRef, ExternalPackageRefCategory
from spdx_tools.spdx.validation.external_package_ref_validator import (
    BOWER_REGEX,
    CPE22TYPE_REGEX,
    CPE23TYPE_REGEX,
    GITOID_REGEX,
    MAVEN_CENTRAL_REGEX,
    NPM_REGEX,
    NUGET_REGEX,
    PURL_REGEX,
    SWH_REGEX,
    validate_external_package_ref,
)
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


@pytest.mark.parametrize(
    "category, reference_type, locator",
    [
        (ExternalPackageRefCategory.SECURITY, "cpe22Type", "cpe:/o:canonical:ubuntu_linux:10.04:-:lts"),
        (ExternalPackageRefCategory.SECURITY, "cpe23Type", "cpe:2.3:o:canonical:ubuntu_linux:10.04:-:lts:*:*:*:*:*"),
        (ExternalPackageRefCategory.SECURITY, "advisory", "https://nvd.nist.gov/vuln/detail/CVE-2020-28498"),
        (ExternalPackageRefCategory.SECURITY, "fix", "https://github.com/indutny/elliptic/commit/441b7428"),
        (
            ExternalPackageRefCategory.SECURITY,
            "url",
            "https://github.com/christianlundkvist/blog/blob/master/2020_05_26_secp256k1_twist_attacks/"
            "secp256k1_twist_attacks.md",
        ),
        (ExternalPackageRefCategory.SECURITY, "swid", "swid:2df9de35-0aff-4a86-ace6-f7dddd1ade4c"),
        (ExternalPackageRefCategory.PACKAGE_MANAGER, "maven-central", "org.apache.tomcat:tomcat:9.0.0.M4"),
        (ExternalPackageRefCategory.PACKAGE_MANAGER, "npm", "http-server@0.3.0"),
        (ExternalPackageRefCategory.PACKAGE_MANAGER, "nuget", "Microsoft.AspNet.MVC/5.0.0"),
        (ExternalPackageRefCategory.PACKAGE_MANAGER, "bower", "modernizr#2.6.2"),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "purl",
            "pkg:docker/debian@sha256:2f04d3d33b6027bb74ecc81397abe780649ec89f1a2af18d7022737d0482cefe",
        ),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "purl",
            "pkg:bitbucket/birkenfeld/pygments-main@244fd47e07d1014f0aed9c",
        ),
        (ExternalPackageRefCategory.PACKAGE_MANAGER, "purl", "pkg:deb/debian/curl@7.50.3-1?arch=i386&distro=jessie"),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "purl",
            "pkg:docker/customer/dockerimage@sha256:244fd47e07d1004f0aed9c?repository_url=gcr.io",
        ),
        (ExternalPackageRefCategory.PACKAGE_MANAGER, "purl", "pkg:gem/jruby-launcher@1.1.2?platform=java"),
        (ExternalPackageRefCategory.PACKAGE_MANAGER, "purl", "pkg:gem/ruby-advisory-db-check@0.12.4"),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "purl",
            "pkg:github/package-url/purl-spec@244fd47e07d1004f0aed9c",
        ),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "purl",
            "pkg:golang/google.golang.org/genproto#googleapis/api/annotations",
        ),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "purl",
            "pkg:maven/org.apache.xmlgraphics/batik-anim@1.9.1?repository_url=repo.spring.io%2Frelease",
        ),
        (ExternalPackageRefCategory.PACKAGE_MANAGER, "purl", "pkg:npm/%40angular/animation@12.3.1"),
        (ExternalPackageRefCategory.PACKAGE_MANAGER, "purl", "pkg:nuget/EnterpriseLibrary.Common@6.0.1304"),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "purl",
            "pkg:rpm/fedora/curl@7.50.3-1.fc25?arch=i386&distro=fedora-25",
        ),
        (ExternalPackageRefCategory.PERSISTENT_ID, "swh", "swh:1:cnt:94a9ed024d3859793618152ea559a168bbcbb5e2"),
        (ExternalPackageRefCategory.PERSISTENT_ID, "swh", "swh:1:dir:d198bc9d7a6bcf6db04f476d29314f157507d505"),
        (ExternalPackageRefCategory.PERSISTENT_ID, "swh", "swh:1:rev:309cf2674ee7a0749978cf8265ab91a60aea0f7d"),
        (ExternalPackageRefCategory.PERSISTENT_ID, "swh", "swh:1:rel:22ece559cc7cc2364edc5e5593d63ae8bd229f9f"),
        (ExternalPackageRefCategory.PERSISTENT_ID, "swh", "swh:1:snp:c7c108084bc0bf3d81436bf980b46e98bd338453"),
        (
            ExternalPackageRefCategory.PERSISTENT_ID,
            "gitoid",
            "gitoid:blob:sha1:261eeb9e9f8b2b4b0d119366dda99c6fd7d35c64",
        ),
        (
            ExternalPackageRefCategory.PERSISTENT_ID,
            "gitoid",
            "gitoid:blob:sha256:3557f7eb43c621c71483743d4b37059bb80933e7f71277c0c3b3846159d1f61c",
        ),
        (ExternalPackageRefCategory.OTHER, "some idstring", "#//string-withOUT!Spaces\\?"),
    ],
)
def test_valid_external_package_ref(category, reference_type, locator):
    external_package_ref = ExternalPackageRef(category, reference_type, locator, "externalPackageRef comment")
    validation_messages: List[ValidationMessage] = validate_external_package_ref(
        external_package_ref, "parent_id", "SPDX-2.3"
    )

    assert validation_messages == []


@pytest.mark.parametrize(
    "category, reference_type, locator, expected_message",
    [
        (
            ExternalPackageRefCategory.SECURITY,
            "cpe22Typo",
            "cpe:/o:canonical:ubuntu_linux:10.04:-:lts",
            "externalPackageRef type in category SECURITY must be one of ['cpe22Type', 'cpe23Type', 'advisory', 'fix'"
            ", 'url', 'swid'], but is: cpe22Typo",
        ),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "nugat",
            "cpe:/o:canonical:ubuntu_linux:10.04:-:lts",
            "externalPackageRef type in category PACKAGE_MANAGER must be one of ['maven-central', 'npm', 'nuget',"
            " 'bower', 'purl'], but is: nugat",
        ),
        (
            ExternalPackageRefCategory.PERSISTENT_ID,
            "git-oid",
            "cpe:/o:canonical:ubuntu_linux:10.04:-:lts",
            "externalPackageRef type in category PERSISTENT_ID must be one of ['swh', 'gitoid'], but is: git-oid",
        ),
    ],
)
def test_invalid_external_package_ref_types(category, reference_type, locator, expected_message):
    external_package_ref = ExternalPackageRef(category, reference_type, locator, "externalPackageRef comment")
    parent_id = "SPDXRef-Package"
    validation_messages: List[ValidationMessage] = validate_external_package_ref(
        external_package_ref, parent_id, "SPDX-2.3"
    )

    expected = ValidationMessage(
        expected_message,
        ValidationContext(
            parent_id=parent_id, element_type=SpdxElementType.EXTERNAL_PACKAGE_REF, full_element=external_package_ref
        ),
    )

    assert validation_messages == [expected]


@pytest.mark.parametrize(
    "category, reference_type, locator, expected_message",
    [
        (
            ExternalPackageRefCategory.SECURITY,
            "cpe22Type",
            "cpe:o:canonical:ubuntu_linux:10.04:-:lts",
            f'externalPackageRef locator of type "cpe22Type" must conform with the regex {CPE22TYPE_REGEX}, but is: '
            f"cpe:o:canonical:ubuntu_linux:10.04:-:lts",
        ),
        (
            ExternalPackageRefCategory.SECURITY,
            "cpe23Type",
            "cpe:2.3:/o:canonical:ubuntu_linux:10.04:-:lts:*:*:*:*:*",
            f'externalPackageRef locator of type "cpe23Type" must conform with the regex {CPE23TYPE_REGEX}, but is: '
            f"cpe:2.3:/o:canonical:ubuntu_linux:10.04:-:lts:*:*:*:*:*",
        ),
        (
            ExternalPackageRefCategory.SECURITY,
            "advisory",
            "http://locatorurl",
            'externalPackageRef locator of type "advisory" must be a valid URL, but is: http://locatorurl',
        ),
        (
            ExternalPackageRefCategory.SECURITY,
            "fix",
            "http://fixurl",
            'externalPackageRef locator of type "fix" must be a valid URL, but is: http://fixurl',
        ),
        (
            ExternalPackageRefCategory.SECURITY,
            "url",
            "http://url",
            'externalPackageRef locator of type "url" must be a valid URL, but is: http://url',
        ),
        (
            ExternalPackageRefCategory.SECURITY,
            "swid",
            "2df9de35-0aff-4a86-ace6-f7dddd1ade4c",
            'externalPackageRef locator of type "swid" must be a valid URI with scheme swid, but is: '
            "2df9de35-0aff-4a86-ace6-f7dddd1ade4c",
        ),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "maven-central",
            "org.apache.tomcat:tomcat:tomcat:9.0.0.M4",
            f'externalPackageRef locator of type "maven-central" must conform with the regex {MAVEN_CENTRAL_REGEX}, '
            f"but is: org.apache.tomcat:tomcat:tomcat:9.0.0.M4",
        ),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "npm",
            "http-server:0.3.0",
            f'externalPackageRef locator of type "npm" must conform with the regex {NPM_REGEX}, '
            f"but is: http-server:0.3.0",
        ),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "nuget",
            "Microsoft.AspNet.MVC@5.0.0",
            f'externalPackageRef locator of type "nuget" must conform with the regex {NUGET_REGEX}, '
            f"but is: Microsoft.AspNet.MVC@5.0.0",
        ),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "bower",
            "modernizr:2.6.2",
            f'externalPackageRef locator of type "bower" must conform with the regex {BOWER_REGEX}, '
            f"but is: modernizr:2.6.2",
        ),
        (
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "purl",
            "pkg:npm@12.3.1",
            f'externalPackageRef locator of type "purl" must conform with the regex {PURL_REGEX}, '
            f"but is: pkg:npm@12.3.1",
        ),
        (
            ExternalPackageRefCategory.PERSISTENT_ID,
            "swh",
            "swh:cnt:94a9ed024d3859793618152ea559a168bbcbb5e2",
            f'externalPackageRef locator of type "swh" must conform with the regex {SWH_REGEX}, '
            f"but is: swh:cnt:94a9ed024d3859793618152ea559a168bbcbb5e2",
        ),
        (
            ExternalPackageRefCategory.PERSISTENT_ID,
            "gitoid",
            "gitoid:blob:sha1:3557f7eb43c621c71483743d4b37059bb80933e7f71277c0c3b3846159d1f61c",
            f'externalPackageRef locator of type "gitoid" must conform with the regex {GITOID_REGEX}, '
            f"but is: gitoid:blob:sha1:3557f7eb43c621c71483743d4b37059bb80933e7f71277c0c3b3846159d1f61c",
        ),
        (
            ExternalPackageRefCategory.PERSISTENT_ID,
            "gitoid",
            "gitoid:blob:sha256:261eeb9e9f8b2b4b0d119366dda99c6fd7d35c64",
            f'externalPackageRef locator of type "gitoid" must conform with the regex {GITOID_REGEX},'
            f" but is: gitoid:blob:sha256:261eeb9e9f8b2b4b0d119366dda99c6fd7d35c64",
        ),
        (
            ExternalPackageRefCategory.OTHER,
            "id string",
            "locator string",
            "externalPackageRef locator in category OTHER must contain no spaces, but is: locator string",
        ),
    ],
)
def test_invalid_external_package_ref_locators(category, reference_type, locator, expected_message):
    external_package_ref = ExternalPackageRef(category, reference_type, locator, "externalPackageRef comment")
    parent_id = "SPDXRef-Package"
    validation_messages: List[ValidationMessage] = validate_external_package_ref(
        external_package_ref, parent_id, "SPDX-2.3"
    )

    expected = ValidationMessage(
        expected_message,
        ValidationContext(
            parent_id=parent_id, element_type=SpdxElementType.EXTERNAL_PACKAGE_REF, full_element=external_package_ref
        ),
    )

    assert validation_messages == [expected]


@pytest.mark.parametrize(
    "category, reference_type, locator",
    [
        (ExternalPackageRefCategory.SECURITY, "advisory", "https://nvd.nist.gov/vuln/detail/CVE-2020-28498"),
        (ExternalPackageRefCategory.SECURITY, "fix", "https://github.com/indutny/elliptic/commit/441b7428"),
        (
            ExternalPackageRefCategory.SECURITY,
            "url",
            "https://github.com/christianlundkvist/blog/blob/master/2020_05_26_secp256k1_twist_attacks/"
            "secp256k1_twist_attacks.md",
        ),
        (ExternalPackageRefCategory.SECURITY, "swid", "swid:2df9de35-0aff-4a86-ace6-f7dddd1ade4c"),
    ],
)
def test_v2_3only_external_package_ref_types(category, reference_type, locator):
    external_package_ref = ExternalPackageRef(category, reference_type, locator, "externalPackageRef comment")
    parent_id = "SPDXRef-Package"
    validation_messages: List[ValidationMessage] = validate_external_package_ref(
        external_package_ref, parent_id, "SPDX-2.2"
    )

    expected = ValidationMessage(
        f'externalPackageRef type "{reference_type}" is not supported in SPDX-2.2',
        ValidationContext(
            parent_id=parent_id, element_type=SpdxElementType.EXTERNAL_PACKAGE_REF, full_element=external_package_ref
        ),
    )

    assert validation_messages == [expected]



#!/usr/bin/env python3

# Copyright (c) 2020 Yash Varshney
# Copyright (c) 2023 spdx contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import sys
from json import JSONDecodeError
from xml.parsers.expat import ExpatError
from xml.sax import SAXParseException

import click
from beartype.typing import List
from yaml.scanner import ScannerError

from spdx_tools.spdx.graph_generation import export_graph_from_document
from spdx_tools.spdx.model import Document
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.parse_anything import parse_file
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document
from spdx_tools.spdx.validation.validation_message import ValidationMessage
from spdx_tools.spdx.writer.tagvalue import tagvalue_writer
from spdx_tools.spdx.writer.write_anything import write_file


@click.command()
@click.option("--infile", "-i", required=True, help="The file containing the document to be validated or converted.")
@click.option(
    "--outfile",
    "-o",
    help="The file to write the converted document to (write a dash for output to stdout or omit for no conversion). "
    "If you add the option --graph to the command the generated graph will be written to this file.",
)
@click.option(
    "--version",
    help='The SPDX version to be used during parsing and validation ("SPDX-2.2" or "SPDX-2.3"). '
    "Will be read from the document if not provided.",
    default=None,
)
@click.option("--novalidation", is_flag=True, help="Don't validate the provided document.")
@click.option(
    "--graph",
    is_flag=True,
    default=False,
    help="Generate a relationship graph from the input file. "
    "The generated graph is saved to the file specified with --outfile. "
    "Note: You need to install the optional dependencies 'networkx' and 'pygraphviz' for this feature.",
)
def main(infile: str, outfile: str, version: str, novalidation: bool, graph: bool):
    """
    CLI-tool for validating SPDX documents and converting between RDF, TAG-VALUE, JSON, YAML and XML formats.
    Formats are determined by the file endings.
    To use, run: 'pyspdxtools --infile <input file name> --outfile <output file name>'
    """
    try:
        document: Document = parse_file(infile)

        if not novalidation:
            if not version:
                version = document.creation_info.spdx_version

            if version not in ["SPDX-2.2", "SPDX-2.3"]:
                logging.error(f"This tool only supports SPDX versions SPDX-2.2 and SPDX-2.3, but got: {version}")
                sys.exit(1)

            validation_messages: List[ValidationMessage] = validate_full_spdx_document(document, version)
            if validation_messages:
                log_string = "\n".join(
                    ["The document is invalid. The following issues have been found:"]
                    + [message.validation_message for message in validation_messages]
                )
                logging.error(log_string)
                sys.exit(1)
            else:
                logging.info("The document is valid.")

        if outfile == "-":
            tagvalue_writer.write_document(document, sys.stdout)

        elif graph:
            try:
                export_graph_from_document(document, outfile)
            except ImportError:
                logging.error(
                    "To be able to draw a relationship graph of the parsed document "
                    "you need to install 'networkx' and 'pygraphviz'. Run 'pip install \".[graph_generation]\"'."
                )
                sys.exit(1)

        elif outfile:
            write_file(document, outfile, validate=False)

    except NotImplementedError as err:
        logging.error(
            err.args[0]
            + "\nPlease note that this project is currently undergoing a major refactoring and therefore missing "
            "a few features which will be added in time (refer to https://github.com/spdx/tools-python/issues "
            "for insights into the current status).\n"
            "In the meantime, please use the current PyPI release version."
        )
        sys.exit(1)

    except SPDXParsingError as err:
        log_string = "\n".join(
            ["There have been issues while parsing the provided document:"]
            + [message for message in err.get_messages()]
        )
        logging.error(log_string)
        sys.exit(1)

    except JSONDecodeError as err:
        logging.error(f"Invalid JSON provided: {err.args[0]}")
        sys.exit(1)

    except ScannerError as err:
        logging.error("Invalid YAML provided: " + "\n".join([str(arg) for arg in err.args]))
        sys.exit(1)

    except ExpatError as err:
        logging.error(f"Invalid XML provided: {err.args[0]}")
        sys.exit(1)

    except SAXParseException as err:
        logging.error(f"Invalid RDF-XML provided: {str(err)}")
        sys.exit(1)

    except FileNotFoundError as err:
        logging.error(f"{err.strerror}: {err.filename}")
        sys.exit(1)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm, FileType, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import parse_list_of_elements
from spdx_tools.spdx.parser.jsonlikedict.file_parser import FileParser


@pytest.mark.parametrize(
    "copyright_text, expected_copyright_text",
    [
        ("Copyright 2008-2010 John Smith", "Copyright 2008-2010 John Smith"),
        ("NOASSERTION", SpdxNoAssertion()),
        ("NONE", SpdxNone()),
    ],
)
def test_parse_file(copyright_text, expected_copyright_text):
    file_parser = FileParser()
    file_dict = {
        "SPDXID": "SPDXRef-File",
        "attributionTexts": ["Some attribution text."],
        "checksums": [
            {"algorithm": "SHA1", "checksumValue": "d6a770ba38583ed4bb4525bd96e50461655d2758"},
            {"algorithm": "MD5", "checksumValue": "624c1abb3664f4b35547e7c73864ad24"},
        ],
        "comment": "The concluded license was taken from the package level that the file was included in.\nThis "
        "information was found in the COPYING.txt file in the xyz directory.",
        "copyrightText": copyright_text,
        "fileContributors": [
            "The Regents of the University of California",
            "Modified by Paul Mundt lethal@linux-sh.org",
            "IBM Corporation",
        ],
        "fileName": "./package/foo.c",
        "fileTypes": ["SOURCE"],
        "licenseComments": "The concluded license was taken from the package level that the file was included in.",
        "licenseConcluded": "(LGPL-2.0-only OR LicenseRef-2)",
        "licenseInfoInFiles": ["GPL-2.0-only", "LicenseRef-2", "NOASSERTION"],
        "noticeText": "Copyright (c) 2001 Aaron Lehmann aaroni@vitelus.com\n\nPermission is hereby granted, free of "
        "charge, to any person obtaining a copy of this software and associated documentation files "
        '(the "Software"), to deal in the Software without restriction, including without limitation '
        "the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of "
        "the Software, and to permit persons to whom the Software is furnished to do so, subject to the "
        "following conditions: \nThe above copyright notice and this permission notice shall be "
        "included in all copies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED "
        '"AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE '
        "WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO "
        "EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER "
        "LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN "
        "CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.",
    }

    file = file_parser.parse_file(file_dict)

    assert file.name == "./package/foo.c"
    assert file.spdx_id == "SPDXRef-File"
    TestCase().assertCountEqual(
        file.checksums,
        [
            Checksum(ChecksumAlgorithm.SHA1, "d6a770ba38583ed4bb4525bd96e50461655d2758"),
            Checksum(ChecksumAlgorithm.MD5, "624c1abb3664f4b35547e7c73864ad24"),
        ],
    )
    assert (
        file.comment
        == "The concluded license was taken from the package level that the file was included in.\nThis information "
        "was found in the COPYING.txt file in the xyz directory."
    )
    assert file.copyright_text == expected_copyright_text
    assert file.file_types == [FileType.SOURCE]
    TestCase().assertCountEqual(
        file.contributors,
        [
            "The Regents of the University of California",
            "Modified by Paul Mundt lethal@linux-sh.org",
            "IBM Corporation",
        ],
    )
    assert file.license_concluded == spdx_licensing.parse("(LGPL-2.0-only OR LicenseRef-2)")
    TestCase().assertCountEqual(
        file.license_info_in_file,
        [spdx_licensing.parse("GPL-2.0-only"), spdx_licensing.parse("LicenseRef-2"), SpdxNoAssertion()],
    )
    assert (
        file.license_comment == "The concluded license was taken from the package level that the file was included in."
    )
    assert file.attribution_texts == ["Some attribution text."]


def test_parse_invalid_files():
    file_parser = FileParser()
    files = [
        {"SPDXID": "SPDXRef-File", "fileName": "Incomplete File"},
        {
            "SPDXID": "SPDXRef-File",
            "attributionTexts": ["Some attribution text."],
            "checksums": [
                {"algorithm": "SHA1", "checksumValue": "d6a770ba38583ed4bb4525bd96e50461655d2758"},
                {"algorithm": "MD5", "checksumValue": "624c1abb3664f4b35547e7c73864ad24"},
            ],
        },
        {
            "SPDXID": "SPDXRef-File",
            "attributionTexts": ["Some attribution text."],
            "checksums": [
                {"algorithm": "SHA1", "checksumValue": "d6a770ba38583ed4bb4525bd96e50461655d2758"},
                {"algorithm": "MD", "checksumValue": "624c1abb3664f4b35547e7c73864ad24"},
            ],
        },
        {"SPDXID": "SPDXRef-File", "fileName": "Incomplete File"},
    ]

    with pytest.raises(SPDXParsingError):
        parse_list_of_elements(files, file_parser.parse_file)


def test_parse_file_types():
    file_parser = FileParser()
    file_types_list = ["OTHER", "APPLICATION"]

    file_types = file_parser.parse_file_types(file_types_list)

    TestCase().assertCountEqual(file_types, [FileType.OTHER, FileType.APPLICATION])


def test_parse_invalid_file_types():
    file_parser = FileParser()
    file_types_list = ["OTHER", "APPLICAON"]

    with pytest.raises(SPDXParsingError) as err:
        file_parser.parse_file_types(file_types_list)

    TestCase().assertCountEqual(
        err.value.get_messages(), ["Error while parsing FileType: ['Invalid FileType: APPLICAON']"]
    )



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import List

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import CreationInfo
from spdx_tools.spdx.validation.actor_validator import validate_actors
from spdx_tools.spdx.validation.external_document_ref_validator import validate_external_document_refs
from spdx_tools.spdx.validation.uri_validators import validate_uri
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_creation_info(creation_info: CreationInfo, spdx_version: str) -> List[ValidationMessage]:
    validation_messages: List[ValidationMessage] = []

    context = ValidationContext(spdx_id=creation_info.spdx_id, element_type=SpdxElementType.DOCUMENT)

    if creation_info.spdx_id != DOCUMENT_SPDX_ID:
        validation_messages.append(
            ValidationMessage(f"spdx_id must be {DOCUMENT_SPDX_ID}, but is: {creation_info.spdx_id}", context)
        )

    if creation_info.data_license != "CC0-1.0":
        validation_messages.append(
            ValidationMessage(f'data_license must be "CC0-1.0", but is: {creation_info.data_license}', context)
        )

    for message in validate_uri(creation_info.document_namespace):
        validation_messages.append(ValidationMessage("document_namespace " + message, context))

    validation_messages.extend(validate_actors(creation_info.creators, creation_info.spdx_id))
    validation_messages.extend(
        validate_external_document_refs(creation_info.external_document_refs, creation_info.spdx_id, spdx_version)
    )

    return validation_messages



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest
from rdflib import RDF, Graph, Literal, URIRef

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.license_expression_writer import add_license_expression_to_graph


def test_add_conjunctive_license_set_to_graph():
    graph = Graph()
    license_expression = spdx_licensing.parse("MIT AND GPL-2.0")

    add_license_expression_to_graph(
        license_expression, graph, URIRef("parentNode"), SPDX_NAMESPACE.licenseConcluded, "https://namespace"
    )

    assert (URIRef("parentNode"), SPDX_NAMESPACE.licenseConcluded, None) in graph
    assert (None, RDF.type, SPDX_NAMESPACE.ConjunctiveLicenseSet) in graph
    assert (None, SPDX_NAMESPACE.member, URIRef("http://spdx.org/licenses/MIT")) in graph
    assert (None, SPDX_NAMESPACE.member, URIRef("http://spdx.org/licenses/GPL-2.0-only")) in graph


def test_add_disjunctive_license_set_to_graph():
    graph = Graph()
    license_expression = spdx_licensing.parse("MIT OR GPL-2.0")

    add_license_expression_to_graph(
        license_expression, graph, URIRef("parentNode"), SPDX_NAMESPACE.licenseConcluded, "https://namespace"
    )

    assert (URIRef("parentNode"), SPDX_NAMESPACE.licenseConcluded, None) in graph
    assert (None, RDF.type, SPDX_NAMESPACE.DisjunctiveLicenseSet) in graph
    assert (None, SPDX_NAMESPACE.member, URIRef("http://spdx.org/licenses/MIT")) in graph
    assert (None, SPDX_NAMESPACE.member, URIRef("http://spdx.org/licenses/GPL-2.0-only")) in graph


@pytest.mark.parametrize(
    "license_with_exception," "expected_triple",
    [
        (
            "MIT WITH openvpn-openssl-exception",
            (URIRef("http://spdx.org/licenses/openvpn-openssl-exception"), RDF.type, SPDX_NAMESPACE.LicenseException),
        ),
        ("MIT WITH unknown-exception", (None, SPDX_NAMESPACE.licenseExceptionId, Literal("unknown-exception"))),
    ],
)
def test_license_exception_to_graph(license_with_exception, expected_triple):
    graph = Graph()
    license_expression = spdx_licensing.parse(license_with_exception)

    add_license_expression_to_graph(
        license_expression, graph, URIRef("parentNode"), SPDX_NAMESPACE.licenseConcluded, "https://namespace"
    )

    assert (URIRef("parentNode"), SPDX_NAMESPACE.licenseConcluded, None) in graph
    assert (None, RDF.type, SPDX_NAMESPACE.WithExceptionOperator) in graph
    assert (None, SPDX_NAMESPACE.member, URIRef("http://spdx.org/licenses/MIT")) in graph
    assert (None, SPDX_NAMESPACE.licenseException, None) in graph
    assert expected_triple in graph



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field
from datetime import datetime

from beartype.typing import List, Optional
from semantic_version import Version

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model import ProfileIdentifierType


@dataclass_with_properties
class CreationInfo:
    spec_version: Version
    created: datetime
    created_by: List[str]  # SPDXID of Agents
    profile: List[ProfileIdentifierType]
    data_license: Optional[str] = "CC0-1.0"
    created_using: List[str] = field(default_factory=list)  # SPDXID of Tools
    comment: Optional[str] = None

    def __init__(
        self,
        spec_version: Version,
        created: datetime,
        created_by: List[str],
        profile: List[ProfileIdentifierType],
        data_license: Optional[str] = "CC0-1.0",
        created_using: List[str] = None,
        comment: Optional[str] = None,
    ):
        created_using = [] if created_using is None else created_using
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum, auto

from beartype.typing import Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model import IntegrityMethod


class HashAlgorithm(Enum):
    BLAKE2B256 = auto()
    BLAKE2B384 = auto()
    BLAKE2B512 = auto()
    BLAKE3 = auto()
    CRYSTALS_KYBER = auto()
    CRYSTALS_DILITHIUM = auto()
    FALCON = auto()
    MD2 = auto()
    MD4 = auto()
    MD5 = auto()
    MD6 = auto()
    OTHER = auto()
    SHA1 = auto()
    SHA224 = auto()
    SHA256 = auto()
    SHA3_224 = auto()
    SHA3_256 = auto()
    SHA3_384 = auto()
    SHA3_512 = auto()
    SHA384 = auto()
    SHA512 = auto()
    SPDXPVCSHA1 = auto()
    SPDXPVCSHA256 = auto()
    SPHINCS_PLUS = auto()


@dataclass_with_properties
class Hash(IntegrityMethod):
    algorithm: HashAlgorithm = None
    hash_value: str = None

    def __init__(self, algorithm: HashAlgorithm, hash_value: str, comment: Optional[str] = None):
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

import pytest
from semantic_version import Version

from spdx_tools.spdx3.model import CreationInfo, ProfileIdentifierType
from tests.spdx3.fixtures import creation_info_fixture
from tests.spdx3.model.model_test_utils import get_property_names


def test_correct_initialization():
    creation_info = creation_info_fixture()

    for property_name in get_property_names(CreationInfo):
        assert getattr(creation_info, property_name) is not None

    assert creation_info.spec_version == Version("3.0.0")
    assert creation_info.created == datetime(2022, 12, 1)
    assert creation_info.created_by == ["https://spdx.test/tools-python/creation_info_created_by"]
    assert creation_info.created_using == ["https://spdx.test/tools-python/creation_info_created_using"]
    assert creation_info.profile == [
        ProfileIdentifierType.CORE,
        ProfileIdentifierType.SOFTWARE,
        ProfileIdentifierType.LICENSING,
    ]
    assert creation_info.data_license == "CC0-1.0"
    assert creation_info.comment == "creationInfoComment"


def test_invalid_initialization():
    with pytest.raises(TypeError) as err:
        CreationInfo("2.3", "2012-01-01", [], "core", 3, [], [])

    assert len(err.value.args[0]) == 5
    for error in err.value.args[0]:
        assert error.startswith("SetterError CreationInfo:")


def test_incomplete_initialization():
    with pytest.raises(TypeError) as err:
        CreationInfo("2.3")

    assert (
        "__init__() missing 3 required positional arguments: 'created', 'created_by', and 'profile'"
        in err.value.args[0]
    )



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest

from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import (
    json_str_to_enum_name,
    parse_field_or_no_assertion,
    parse_field_or_no_assertion_or_none,
)


def test_json_str_to_enum():
    json_str = "BLAKE2b-256"

    enum_name = json_str_to_enum_name(json_str)

    assert enum_name == "BLAKE2B_256"


@pytest.mark.parametrize("invalid_json_str,expected_message", [(5, ["Type for enum must be str not int"])])
def test_invalid_json_str_to_enum(invalid_json_str, expected_message):
    with pytest.raises(SPDXParsingError) as err:
        json_str_to_enum_name(invalid_json_str)

    TestCase().assertCountEqual(err.value.get_messages(), expected_message)


@pytest.mark.parametrize("input_str,expected_type", [("NOASSERTION", SpdxNoAssertion), ("example string", str)])
def test_parse_field_or_no_assertion(input_str, expected_type):
    resulting_value = parse_field_or_no_assertion(input_str, lambda x: x)

    assert isinstance(resulting_value, expected_type)


@pytest.mark.parametrize(
    "input_str,expected_type", [("NOASSERTION", SpdxNoAssertion), ("NONE", SpdxNone), ("example string", str)]
)
def test_parse_field_or_no_assertion_or_none(input_str, expected_type):
    resulting_value = parse_field_or_no_assertion_or_none(input_str, lambda x: x)

    assert isinstance(resulting_value, expected_type)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Type

from spdx_tools.spdx.jsonschema.checksum_converter import ChecksumConverter
from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.external_document_ref_properties import ExternalDocumentRefProperty
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.model import Document, ExternalDocumentRef


class ExternalDocumentRefConverter(TypedConverter[ExternalDocumentRef]):
    checksum_converter: ChecksumConverter

    def __init__(self):
        self.checksum_converter = ChecksumConverter()

    def _get_property_value(
        self,
        external_document_ref: ExternalDocumentRef,
        external_document_ref_property: ExternalDocumentRefProperty,
        _document: Document = None,
    ) -> Any:
        if external_document_ref_property == ExternalDocumentRefProperty.EXTERNAL_DOCUMENT_ID:
            return external_document_ref.document_ref_id
        elif external_document_ref_property == ExternalDocumentRefProperty.SPDX_DOCUMENT:
            return external_document_ref.document_uri
        elif external_document_ref_property == ExternalDocumentRefProperty.CHECKSUM:
            return self.checksum_converter.convert(external_document_ref.checksum)

    def get_json_type(self) -> Type[JsonProperty]:
        return ExternalDocumentRefProperty

    def get_data_model_type(self) -> Type[ExternalDocumentRef]:
        return ExternalDocumentRef



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import os
from datetime import datetime
from unittest.mock import MagicMock, call, mock_open, patch

import pytest

from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    Checksum,
    ChecksumAlgorithm,
    CreationInfo,
    Document,
    File,
    Package,
    Relationship,
    RelationshipType,
    Snippet,
)
from spdx_tools.spdx.parser.tagvalue import tagvalue_parser
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer import write_document, write_document_to_file
from tests.spdx.fixtures import checksum_fixture, document_fixture


@pytest.fixture
def temporary_file_path() -> str:
    temporary_file_path = "temp_test_tag_value_writer_output.spdx"
    yield temporary_file_path
    os.remove(temporary_file_path)


def test_write_tag_value(temporary_file_path: str):
    document = document_fixture()

    write_document_to_file(document, temporary_file_path, False)

    parsed_document = tagvalue_parser.parse_from_file(temporary_file_path)

    assert parsed_document == document


def test_correct_order_of_elements():
    packages = [
        Package(name="Test Package A", spdx_id="SPDXRef-Package-A", download_location=""),
        Package(name="Test Package B", spdx_id="SPDXRef-Package-B", download_location=""),
    ]
    files = [
        File(name="Test File A", spdx_id="SPDXRef-File-A", checksums=[checksum_fixture()]),
        File(name="Test File B", spdx_id="SPDXRef-File-B", checksums=[checksum_fixture()]),
    ]
    snippets = [
        Snippet(spdx_id="SPDXRef-Snippet-A", file_spdx_id="DocumentRef-External:SPDXRef-File", byte_range=(1, 2)),
        Snippet(spdx_id="SPDXRef-Snippet-B", file_spdx_id="SPDXRef-File-A", byte_range=(1, 2)),
        Snippet(spdx_id="SPDXRef-Snippet-C", file_spdx_id="SPDXRef-File-B", byte_range=(3, 4)),
    ]
    relationships = [Relationship("SPDXRef-Package-A", RelationshipType.CONTAINS, "SPDXRef-File-B")]
    document = document_fixture(
        files=files,
        packages=packages,
        snippets=snippets,
        relationships=relationships,
        annotations=[],
        extracted_licensing_info=[],
    )

    mock: MagicMock = mock_open()
    with patch(f"{__name__}.open", mock, create=True):
        with open("foo", "w") as file:
            write_document(document, file)

    mock.assert_called_once_with("foo", "w")
    handle = mock()
    handle.write.assert_has_calls(
        [
            call("## Document Information\n"),
            call("SPDXVersion: SPDX-2.3\n"),
            call("DataLicense: CC0-1.0\n"),
            call("SPDXID: SPDXRef-DOCUMENT\n"),
            call("DocumentName: documentName\n"),
            call("DocumentNamespace: https://some.namespace\n"),
            call("DocumentComment: documentComment\n"),
            call("\n## External Document References\n"),
            call(
                "ExternalDocumentRef: DocumentRef-external https://namespace.com "
                "SHA1: 71c4025dd9897b364f3ebbb42c484ff43d00791c\n"
            ),
            call("\n"),
            call("## Creation Information\n"),
            call("LicenseListVersion: 3.19\n"),
            call("Creator: Person: creatorName (some@mail.com)\n"),
            call("Created: 2022-12-01T00:00:00Z\n"),
            call("CreatorComment: creatorComment\n"),
            call("\n"),
            call("## Snippet Information\n"),
            call("SnippetSPDXID: SPDXRef-Snippet-A\n"),
            call("SnippetFromFileSPDXID: DocumentRef-External:SPDXRef-File\n"),
            call("SnippetByteRange: 1:2\n"),
            call("\n"),
            call("## File Information\n"),
            call("FileName: Test File A\n"),
            call("SPDXID: SPDXRef-File-A\n"),
            call("FileChecksum: SHA1: 71c4025dd9897b364f3ebbb42c484ff43d00791c\n"),
            call("\n"),
            call("## Snippet Information\n"),
            call("SnippetSPDXID: SPDXRef-Snippet-B\n"),
            call("SnippetFromFileSPDXID: SPDXRef-File-A\n"),
            call("SnippetByteRange: 1:2\n"),
            call("\n"),
            call("## Package Information\n"),
            call("PackageName: Test Package A\n"),
            call("SPDXID: SPDXRef-Package-A\n"),
            call("PackageDownloadLocation: \n"),
            call("FilesAnalyzed: true\n"),
            call("\n"),
            call("## File Information\n"),
            call("FileName: Test File B\n"),
            call("SPDXID: SPDXRef-File-B\n"),
            call("FileChecksum: SHA1: 71c4025dd9897b364f3ebbb42c484ff43d00791c\n"),
            call("\n"),
            call("## Snippet Information\n"),
            call("SnippetSPDXID: SPDXRef-Snippet-C\n"),
            call("SnippetFromFileSPDXID: SPDXRef-File-B\n"),
            call("SnippetByteRange: 3:4\n"),
            call("\n"),
            call("## Package Information\n"),
            call("PackageName: Test Package B\n"),
            call("SPDXID: SPDXRef-Package-B\n"),
            call("PackageDownloadLocation: \n"),
            call("FilesAnalyzed: true\n"),
            call("\n"),
            call("\n"),
        ]
    )


def test_same_file_in_multiple_packages():
    creation_info = CreationInfo(
        spdx_version="SPDX-2.3",
        spdx_id="SPDXRef-DOCUMENT",
        data_license="CC0-1.0",
        name="SPDX Lite Document",
        document_namespace="https://test.namespace.com",
        creators=[Actor(ActorType.PERSON, "John Doe")],
        created=datetime(2023, 3, 14, 8, 49),
    )
    package_a = Package(
        name="Example package A",
        spdx_id="SPDXRef-Package-A",
        download_location="https://download.com",
    )
    package_b = Package(
        name="Example package B",
        spdx_id="SPDXRef-Package-B",
        download_location="https://download.com",
    )
    file = File(
        name="Example file",
        spdx_id="SPDXRef-File",
        checksums=[Checksum(ChecksumAlgorithm.SHA1, "2fd4e1c67a2d28fced849ee1bb76e7391b93eb12")],
    )

    relationships = [
        Relationship("SPDXRef-DOCUMENT", RelationshipType.DESCRIBES, "SPDXRef-Package-A"),
        Relationship("SPDXRef-DOCUMENT", RelationshipType.DESCRIBES, "SPDXRef-Package-B"),
        Relationship("SPDXRef-Package-A", RelationshipType.CONTAINS, "SPDXRef-File"),
        Relationship("SPDXRef-Package-B", RelationshipType.CONTAINS, "SPDXRef-File"),
    ]
    document = Document(
        creation_info=creation_info,
        packages=[package_a, package_b],
        files=[file],
        relationships=relationships,
    )
    mock: MagicMock = mock_open()
    with patch(f"{__name__}.open", mock, create=True):
        with open("foo", "w") as file:
            write_document(document, file)

    mock.assert_called_once_with("foo", "w")
    handle = mock()
    handle.write.assert_has_calls(
        [
            call("## Document Information\n"),
            call("SPDXVersion: SPDX-2.3\n"),
            call("DataLicense: CC0-1.0\n"),
            call("SPDXID: SPDXRef-DOCUMENT\n"),
            call("DocumentName: SPDX Lite Document\n"),
            call("DocumentNamespace: https://test.namespace.com\n"),
            call("\n"),
            call("## Creation Information\n"),
            call("Creator: Person: John Doe\n"),
            call("Created: 2023-03-14T08:49:00Z\n"),
            call("\n"),
            call("## Package Information\n"),
            call("PackageName: Example package A\n"),
            call("SPDXID: SPDXRef-Package-A\n"),
            call("PackageDownloadLocation: https://download.com\n"),
            call("FilesAnalyzed: true\n"),
            call("\n"),
            call("## File Information\n"),
            call("FileName: Example file\n"),
            call("SPDXID: SPDXRef-File\n"),
            call("FileChecksum: SHA1: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12\n"),
            call("\n"),
            call("## Package Information\n"),
            call("PackageName: Example package B\n"),
            call("SPDXID: SPDXRef-Package-B\n"),
            call("PackageDownloadLocation: https://download.com\n"),
            call("FilesAnalyzed: true\n"),
            call("\n"),
            call("## Relationships\n"),
            call("Relationship: SPDXRef-DOCUMENT DESCRIBES SPDXRef-Package-A\n"),
            call("Relationship: SPDXRef-DOCUMENT DESCRIBES SPDXRef-Package-B\n"),
            call("Relationship: SPDXRef-Package-B CONTAINS SPDXRef-File\n"),
            call("\n"),
        ]
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from unittest import TestCase

import pytest

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import ExternalPackageRef, ExternalPackageRefCategory, PackagePurpose, SpdxNone
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.tagvalue.parser import Parser
from tests.spdx.parser.tagvalue.test_creation_info_parser import DOCUMENT_STR


def test_parse_package():
    parser = Parser()
    package_str = "\n".join(
        [
            "PackageName: Test",
            "SPDXID: SPDXRef-Package",
            "PackageVersion: 1:22.36.1-8+deb11u1",
            "PackageDownloadLocation: http://example.com/test",
            "FilesAnalyzed: true",
            "PackageSummary: <text>Test package</text>",
            "PackageSourceInfo: <text>Version 1.0 of test</text>",
            "PackageFileName: test-1.0.zip",
            "PackageSupplier: Organization:ACME",
            "PackageOriginator: Organization:ACME",
            "PackageChecksum: SHA1: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12",
            "PackageVerificationCode: 4e3211c67a2d28fced849ee1bb76e7391b93feba (something.rdf, something.txt)",
            "PackageDescription: <text>A package.</text>",
            "PackageComment: <text>Comment on the package.</text>",
            "PackageCopyrightText: <text> Copyright 2014 Acme Inc.</text>",
            "PackageLicenseDeclared: Apache-2.0",
            "PackageLicenseConcluded: (LicenseRef-2.0 and Apache-2.0)",
            "PackageLicenseInfoFromFiles: Apache-1.0",
            "PackageLicenseInfoFromFiles: Apache-2.0",
            "PackageLicenseInfoFromFiles: NONE",
            "PackageLicenseComments: <text>License Comments</text>",
            "ExternalRef: SECURITY cpe23Type cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*:*:*:*:",
            "ExternalRefComment: <text>Some comment about the package.</text>",
            "ExternalRef: OTHER LocationRef-acmeforge acmecorp/acmenator/4.1.3-alpha",
            "PrimaryPackagePurpose: OPERATING-SYSTEM",
            "BuiltDate: 2020-01-01T12:00:00Z",
            "ReleaseDate: 2021-01-01T12:00:00Z",
            "ValidUntilDate: 2022-01-01T12:00:00Z",
        ]
    )
    document = parser.parse("\n".join([DOCUMENT_STR, package_str]))
    assert document is not None
    package = document.packages[0]
    assert package.name == "Test"
    assert package.spdx_id == "SPDXRef-Package"
    assert package.version == "1:22.36.1-8+deb11u1"
    assert len(package.license_info_from_files) == 3
    TestCase().assertCountEqual(
        package.license_info_from_files,
        [spdx_licensing.parse("Apache-1.0"), spdx_licensing.parse("Apache-2.0"), SpdxNone()],
    )
    assert package.license_concluded == spdx_licensing.parse("LicenseRef-2.0 AND Apache-2.0")
    assert package.files_analyzed is True
    assert package.comment == "Comment on the package."
    assert len(package.external_references) == 2
    TestCase().assertCountEqual(
        package.external_references,
        [
            ExternalPackageRef(
                ExternalPackageRefCategory.SECURITY,
                "cpe23Type",
                "cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*:*:*:*:",
                "Some comment about the package.",
            ),
            ExternalPackageRef(
                ExternalPackageRefCategory.OTHER, "LocationRef-acmeforge", "acmecorp/acmenator/4.1.3-alpha"
            ),
        ],
    )
    assert package.primary_package_purpose == PackagePurpose.OPERATING_SYSTEM
    assert package.built_date == datetime(2020, 1, 1, 12)
    assert package.release_date == datetime(2021, 1, 1, 12)
    assert package.valid_until_date == datetime(2022, 1, 1, 12)


@pytest.mark.parametrize(
    "package_str, expected_message",
    [
        (
            "PackageDownloadLocation: SPDXRef-Package",
            "Element Package is not the current element in scope, probably the expected "
            "tag to start the element (PackageName) is missing. Line: 1",
        ),
        (
            "PackageName: TestPackage",
            r"__init__() missing 2 required positional arguments: 'spdx_id' and 'download_location'",
        ),
        (
            "PackageName: TestPackage\nPackageCopyrightText:This is a copyright\n"
            "PackageCopyrightText:MultipleCopyright",
            "Error while parsing Package: ['Multiple values for PackageCopyrightText " "found. Line: 3']",
        ),
        (
            "PackageName: TestPackage\nExternalRef: reference locator",
            (
                "Error while parsing Package: [\"Couldn't split PackageExternalRef in category, "
                'reference_type and locator. Line: 2"]'
            ),
        ),
        (
            "PackageName: TestPackage\nExternalRef: category reference locator",
            "Error while parsing Package: ['Invalid ExternalPackageRefCategory: " "category. Line: 2']",
        ),
        (
            f"SPDXID:{DOCUMENT_SPDX_ID}\nPackageName: TestPackage\nSPDXID:SPDXRef-Package\n"
            "PackageDownloadLocation: download.com\nPackageVerificationCode: category reference locator",
            "Error while parsing Package: ['Error while parsing PackageVerificationCode: "
            "Value did not match expected format. Line: 5']",
        ),
        (
            "PackageName: TestPackage\nBuiltDate: 2012\nValidUntilDate:202-11-02T00:00",
            "Error while parsing Package: ['Error while parsing BuiltDate: Token did not "
            "match specified grammar rule. Line: 2', 'Error while parsing "
            "ValidUntilDate: Token did not match specified grammar rule. Line: 3']",
        ),
        (
            f"SPDXID:{DOCUMENT_SPDX_ID}\nPackageName: TestPackage\nSPDXID:SPDXRef-Package\n"
            "PackageDownloadLocation: download.com\nFilesAnalyzed: FALSE",
            "Error while parsing Package: "
            '[\'The value of FilesAnalyzed must be either "true" or "false", but is: FALSE\']',
        ),
    ],
)
def test_parse_invalid_package(package_str, expected_message):
    parser = Parser()

    with pytest.raises(SPDXParsingError) as err:
        parser.parse(package_str)

    assert expected_message in err.value.get_messages()[0]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import sys

from spdx_tools.spdx3.bump_from_spdx2.spdx_document import bump_spdx_document
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx3.writer.console.payload_writer import write_payload
from spdx_tools.spdx.model.actor import ActorType
from spdx_tools.spdx.model.document import Document as Spdx2_Document
from tests.spdx.fixtures import actor_fixture, annotation_fixture, creation_info_fixture, document_fixture


def test_bump_spdx_document():
    spdx2_document: Spdx2_Document = document_fixture()
    spdx2_document.creation_info.creators.append(actor_fixture(ActorType.TOOL, "tool_name", None))
    document_namespace = document_fixture().creation_info.document_namespace

    payload: Payload = bump_spdx_document(spdx2_document)

    write_payload(payload, sys.stdout)

    assert f"{document_namespace}#SPDXRef-Package" in payload.get_full_map()
    assert len(payload.get_full_map()) == 11

    # this is more of a temporary test to make sure the dates don't get messed up again
    assert (
        payload.get_element("#".join([document_namespace, "SPDXRef-DOCUMENT"])).creation_info.created
        == creation_info_fixture().created
    )
    assert (
        payload.get_element("#".join([document_namespace, "SPDXRef-Annotation-0"])).creation_info.created
        == annotation_fixture().annotation_date
    )
